import math
from typing import Optional

import pytest
from hypothesis import given
from hypothesis import strategies as st
from typing_extensions import assert_type

from circuitmatter import tlv

# Test TLV encoding using examples from spec

# Type and Value
#  Encoding (hex)
# Boolean false
#  08
# Boolean true
#  09


class Bool(tlv.Structure):
    b = tlv.BoolMember(0)


class TestBool:
    def test_bool_false_decode(self):
        s = Bool.decode(b"\x15\x28\x00\x18")
        assert str(s) == "{\n  b = false\n}"
        assert s.b is False

    def test_bool_true_decode(self):
        s = Bool.decode(b"\x15\x29\x00\x18")
        assert str(s) == "{\n  b = true\n}"
        assert s.b is True

    def test_bool_false_encode(self):
        s = Bool()
        s.b = False
        assert s.encode().tobytes() == b"\x15\x28\x00\x18"

    def test_bool_true_encode(self):
        s = Bool()
        s.b = True
        assert s.encode().tobytes() == b"\x15\x29\x00\x18"


class SignedIntOneOctet(tlv.Structure):
    i = tlv.NumberMember(0, "b")


class SignedIntTwoOctet(tlv.Structure):
    i = tlv.NumberMember(0, "h")


class SignedIntFourOctet(tlv.Structure):
    i = tlv.NumberMember(0, "i")


class SignedIntEightOctet(tlv.Structure):
    i = tlv.NumberMember(0, "q")


# Signed Integer, 1-octet, value 42
#  00 2a
# Signed Integer, 1-octet, value -17
#  00 ef
# Signed Integer, 2-octet, value 42
#  01 2a 00
# Signed Integer, 4-octet, value -170000
#  02 f0 67 fd ff
# Signed Integer, 8-octet, value 40000000000
#  03 00 90 2f 50 09 00 00 00
class TestSignedInt:
    def test_signed_int_42_decode(self):
        s = SignedIntOneOctet.decode(b"\x15\x20\x00\x2a")
        assert str(s) == "{\n  i = 42\n}"
        assert s.i == 42

    def test_signed_int_negative_17_decode(self):
        s = SignedIntOneOctet.decode(b"\x15\x20\x00\xef")
        assert str(s) == "{\n  i = -17\n}"
        assert s.i == -17

    def test_signed_int_42_two_octet_decode(self):
        s = SignedIntTwoOctet.decode(b"\x15\x21\x00\x2a\x00")
        assert str(s) == "{\n  i = 42\n}"
        assert s.i == 42

    def test_signed_int_negative_170000_decode(self):
        s = SignedIntFourOctet.decode(b"\x15\x22\x00\xf0\x67\xfd\xff")
        assert str(s) == "{\n  i = -170000\n}"
        assert s.i == -170000

    def test_signed_int_40000000000_decode(self):
        s = SignedIntEightOctet.decode(b"\x15\x23\x00\x00\x90\x2f\x50\x09\x00\x00\x00")
        assert str(s) == "{\n  i = 40000000000\n}"
        assert s.i == 40000000000

    def test_signed_int_42_encode(self):
        s = SignedIntOneOctet()
        s.i = 42
        assert s.encode().tobytes() == b"\x15\x20\x00\x2a\x18"

    def test_signed_int_negative_17_encode(self):
        s = SignedIntOneOctet()
        s.i = -17
        assert s.encode().tobytes() == b"\x15\x20\x00\xef\x18"

    def test_signed_int_42_two_octet_encode(self):
        s = SignedIntTwoOctet()
        s.i = 42
        assert s.encode().tobytes() == b"\x15\x20\x00\x2a\x18"

    def test_signed_int_negative_170000_encode(self):
        s = SignedIntFourOctet()
        s.i = -170000
        assert s.encode().tobytes() == b"\x15\x22\x00\xf0\x67\xfd\xff\x18"

    def test_signed_int_40000000000_encode(self):
        s = SignedIntEightOctet()
        s.i = 40000000000
        assert (
            s.encode().tobytes() == b"\x15\x23\x00\x00\x90\x2f\x50\x09\x00\x00\x00\x18"
        )

    @pytest.mark.parametrize(
        "octets,lower,upper",
        [
            (1, -128, 127),
            (2, -32_768, 32_767),
            (4, -2_147_483_648, 2_147_483_647),
            (8, -9_223_372_036_854_775_808, 9_223_372_036_854_775_807),
        ],
    )
    def test_bounds_checks(self, octets, lower, upper):
        class SignedIntStruct(tlv.Structure):
            i = tlv.IntMember(None, signed=True, octets=octets)

        s = SignedIntStruct()

        with pytest.raises(ValueError):
            s.i = lower - 1

        with pytest.raises(ValueError):
            s.i = upper + 1

        s.i = lower
        s.i = upper


class UnsignedIntOneOctet(tlv.Structure):
    i = tlv.NumberMember(0, "B")


# Unsigned Integer, 1-octet, value 42U
#  04 2a
class TestUnsignedInt:
    def test_unsigned_int_42_decode(self):
        s = UnsignedIntOneOctet.decode(b"\x15\x24\x00\x2a\x18")
        assert str(s) == "{\n  i = 42U\n}"
        assert s.i == 42

    def test_unsigned_int_42_encode(self):
        s = UnsignedIntOneOctet()
        s.i = 42
        assert s.encode().tobytes() == b"\x15\x24\x00\x2a\x18"

    @pytest.mark.parametrize(
        "octets,lower,upper",
        [
            (1, 0, 255),
            (2, 0, 65_535),
            (4, 0, 4_294_967_295),
            (8, 0, 18_446_744_073_709_551_615),
        ],
    )
    def test_bounds_checks(self, octets, lower, upper):
        class UnsignedIntStruct(tlv.Structure):
            i = tlv.IntMember(None, signed=False, octets=octets)

        s = UnsignedIntStruct()

        with pytest.raises(ValueError):
            s.i = lower - 1

        with pytest.raises(ValueError):
            s.i = upper + 1

        s.i = lower
        s.i = upper

    @given(v=st.integers(min_value=0, max_value=255))
    def test_roundtrip(self, v: int):
        s = UnsignedIntOneOctet()
        s.i = v
        buffer = s.encode().tobytes()

        s2 = UnsignedIntOneOctet.decode(buffer)

        assert s2.i == s.i
        assert str(s2) == str(s)

    def test_nullability(self):
        class Struct(tlv.Structure):
            i = tlv.IntMember(None)
            ni = tlv.IntMember(None, nullable=True)

        s = Struct()
        assert_type(s.i, int)
        assert_type(s.ni, Optional[int])

        s.ni = None
        assert s.ni is None

        with pytest.raises(ValueError):
            s.i = None


# UTF-8 String, 1-octet length, "Hello!"
#  0c 06 48 65 6c 6c 6f 21
# UTF-8 String, 1-octet length, "Tschüs"
#  0c 07 54 73 63 68 c3 bc 73
class UTF8StringOneOctet(tlv.Structure):
    s = tlv.UTF8StringMember(0, 16)


class TestUTF8String:
    def test_utf8_string_hello_decode(self):
        s = UTF8StringOneOctet.decode(b"\x15\x2c\x00\x06Hello!")
        assert str(s) == '{\n  s = "Hello!"\n}'
        assert s.s == "Hello!"

    def test_utf8_string_tschs_decode(self):
        s = UTF8StringOneOctet.decode(b"\x15\x2c\x00\x07Tsch\xc3\xbcs")
        assert str(s) == '{\n  s = "Tschüs"\n}'
        assert s.s == "Tschüs"

    def test_utf8_string_hello_encode(self):
        s = UTF8StringOneOctet()
        s.s = "Hello!"
        assert s.encode().tobytes() == b"\x15\x2c\x00\x06Hello!\x18"

    def test_utf8_string_tschs_encode(self):
        s = UTF8StringOneOctet()
        s.s = "Tschüs"
        assert s.encode().tobytes() == b"\x15\x2c\x00\x07Tsch\xc3\xbcs\x18"

    @given(v=st.text(max_size=4))
    def test_roundtrip(self, v: str):
        s = UTF8StringOneOctet()
        print(len(v))
        s.s = v
        buffer = s.encode().tobytes()

        s2 = UTF8StringOneOctet.decode(buffer)

        assert s2.s == s.s
        assert str(s2) == str(s)


# Octet String, 1-octet length, octets 00 01 02 03 04
# encoded: 10 05 00 01 02 03 04
class OctetStringOneOctet(tlv.Structure):
    s = tlv.OctetStringMember(0, 16)


class TestOctetString:
    def test_octet_string_decode(self):
        s = OctetStringOneOctet.decode(b"\x15\x30\x00\x05\x00\x01\x02\x03\x04\x18")
        assert str(s) == "{\n  s = 00 01 02 03 04\n}"
        assert s.s == b"\x00\x01\x02\x03\x04"

    def test_octet_string_encode(self):
        s = OctetStringOneOctet()
        s.s = b"\x00\x01\x02\x03\x04"
        assert s.encode().tobytes() == b"\x15\x30\x00\x05\x00\x01\x02\x03\x04\x18"

    @given(v=st.binary(max_size=16))
    def test_roundtrip(self, v: bytes):
        s = OctetStringOneOctet()
        s.s = v
        buffer = s.encode().tobytes()

        s2 = OctetStringOneOctet.decode(buffer)

        assert s2.s == s.s
        assert str(s2) == str(s)


# Null
#  14


class Null(tlv.Structure):
    n = tlv.BoolMember(0, nullable=True)


class NotNull(tlv.Structure):
    n = tlv.BoolMember(0, nullable=True)
    b = tlv.BoolMember(1)


class TestNull:
    def test_null_decode(self):
        s = Null.decode(b"\x15\x34\x00\x18")
        assert str(s) == "{\n  n = null\n}"
        assert s.n is None

    def test_null_encode(self):
        s = Null()
        s.n = None
        assert s.encode().tobytes() == b"\x15\x34\x00\x18"

    def test_nullable(self):
        s = NotNull()

        assert_type(s.b, bool)
        with pytest.raises(ValueError):
            s.b = None  # type: ignore  # testing runtime behaviour


# Single precision floating point 0.0
#  0a 00 00 00 00
# Single precision floating point (1.0 / 3.0)
#  0a ab aa aa 3e
# Single precision floating point 17.9
#  0a 33 33 8f 41
# Single precision floating point infinity (∞)
#  0a 00 00 80 7f
# Single precision floating point negative infinity
#  0a 00 00 80 ff
# (-∞)
# Double precision floating point 0.0
#  0b 00 00 00 00 00 00 00 00
# Double precision floating point (1.0 / 3.0)
#  0b 55 55 55 55 55 55 d5 3f
# Double precision floating point 17.9
#  0b 66 66 66 66 66 e6 31 40
# Double precision floating point infinity (∞)
#  0b 00 00 00 00 00 00 f0 7f
# Double precision floating point negative infinity 0b 00 00 00 00 00 00 f0 ff
# (-∞)
class FloatSingle(tlv.Structure):
    f = tlv.FloatMember(0)


class FloatDouble(tlv.Structure):
    f = tlv.FloatMember(0, octets=8)


class TestFloatSingle:
    def test_precision_float_0_0_decode(self):
        s = FloatSingle.decode(b"\x15\x2a\x00\x00\x00\x00\x00\x18")
        assert str(s) == "{\n  f = 0.0\n}"
        assert s.f == 0.0

    def test_precision_float_1_3_decode(self):
        s = FloatSingle.decode(b"\x15\x2a\x00\xab\xaa\xaa\x3e\x18")
        # assert str(s) == "{\n  f = 0.3333333432674408\n}"
        f = s.f
        assert math.isclose(f, 1.0 / 3.0, rel_tol=1e-06)

    def test_precision_float_17_9_decode(self):
        s = FloatSingle.decode(b"\x15\x2a\x00\x33\x33\x8f\x41\x18")
        assert str(s) == "{\n  f = 17.899999618530273\n}"
        assert math.isclose(s.f, 17.9, rel_tol=1e-06)

    def test_precision_float_infinity_decode(self):
        s = FloatSingle.decode(b"\x15\x2a\x00\x00\x00\x80\x7f\x18")
        assert str(s) == "{\n  f = inf\n}"
        assert math.isinf(s.f)

    def test_precision_float_negative_infinity_decode(self):
        s = FloatSingle.decode(b"\x15\x2a\x00\x00\x00\x80\xff\x18")
        assert str(s) == "{\n  f = -inf\n}"
        assert math.isinf(s.f)

    def test_precision_float_0_0_encode(self):
        s = FloatSingle()
        s.f = 0.0
        assert s.encode().tobytes() == b"\x15\x2a\x00\x00\x00\x00\x00\x18"

    def test_precision_float_1_3_encode(self):
        s = FloatSingle()
        s.f = 1.0 / 3.0
        assert s.encode().tobytes() == b"\x15\x2a\x00\xab\xaa\xaa\x3e\x18"

    def test_precision_float_17_9_encode(self):
        s = FloatSingle()
        s.f = 17.9
        assert s.encode().tobytes() == b"\x15\x2a\x00\x33\x33\x8f\x41\x18"

    def test_precision_float_infinity_encode(self):
        s = FloatSingle()
        s.f = float("inf")
        assert s.encode().tobytes() == b"\x15\x2a\x00\x00\x00\x80\x7f\x18"

    def test_precision_float_negative_infinity_encode(self):
        s = FloatSingle()
        s.f = float("-inf")
        assert s.encode().tobytes() == b"\x15\x2a\x00\x00\x00\x80\xff\x18"

    @given(v=...)
    def test_roundtrip_double(self, v: float):
        s = FloatDouble()
        s.f = v
        buffer = s.encode().tobytes()

        s2 = FloatDouble.decode(buffer)

        assert (
            (math.isnan(s.f) and math.isnan(s2.f))
            or (s.f > 3.4028235e38 and s2.f == float("inf"))
            or (s.f < -3.4028235e38 and s2.f == float("-inf"))
            or math.isclose(s2.f, s.f, rel_tol=1e-7, abs_tol=1e-9)
        )

    @given(
        v=st.floats(
            # encoding to LE float32 raises OverflowError outside these ranges
            # TODO: should we raise ValueError with a bounds check or encode -inf/inf?
            min_value=(2**-126),
            max_value=(2 - 2**-23) * 2**127,
        ),
    )
    def test_roundtrip_single(self, v: float):
        s = FloatSingle()
        s.f = v
        buffer = s.encode().tobytes()
        print("Buffer", buffer.hex(" "))

        s2 = FloatSingle.decode(buffer)

        assert (math.isnan(s.f) and math.isnan(s2.f)) or math.isclose(
            s2.f, s.f, rel_tol=1e-7, abs_tol=1e-9
        )


class TestFloatDouble:
    def test_precision_float_0_0_decode(self):
        s = FloatDouble.decode(b"\x15\x2b\x00\x00\x00\x00\x00\x00\x00\x00\x00")
        assert str(s) == "{\n  f = 0.0\n}"
        assert s.f == 0.0

    def test_precision_float_1_3_decode(self):
        s = FloatDouble.decode(b"\x15\x2b\x00\x55\x55\x55\x55\x55\x55\xd5\x3f")
        # assert str(s) == "{\n  f = 0.3333333333333333\n}"
        f = s.f
        assert math.isclose(f, 1.0 / 3.0, rel_tol=1e-06)

    def test_precision_float_17_9_decode(self):
        s = FloatDouble.decode(b"\x15\x2b\x00\x66\x66\x66\x66\x66\xe6\x31\x40")
        assert str(s) == "{\n  f = 17.9\n}"
        assert math.isclose(s.f, 17.9, rel_tol=1e-06)

    def test_precision_float_infinity_decode(self):
        s = FloatDouble.decode(b"\x15\x2b\x00\x00\x00\x00\x00\x00\x00\xf0\x7f")
        assert str(s) == "{\n  f = inf\n}"
        assert math.isinf(s.f)

    def test_precision_float_negative_infinity_decode(self):
        s = FloatDouble.decode(b"\x15\x2b\x00\x00\x00\x00\x00\x00\x00\xf0\xff")
        assert str(s) == "{\n  f = -inf\n}"
        assert math.isinf(s.f)

    def test_precision_float_0_0_encode(self):
        s = FloatDouble()
        s.f = 0.0
        assert (
            s.encode().tobytes() == b"\x15\x2b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x18"
        )

    def test_precision_float_1_3_encode(self):
        s = FloatDouble()
        s.f = 1.0 / 3.0
        assert (
            s.encode().tobytes() == b"\x15\x2b\x00\x55\x55\x55\x55\x55\x55\xd5\x3f\x18"
        )

    def test_precision_float_17_9_encode(self):
        s = FloatDouble()
        s.f = 17.9
        assert (
            s.encode().tobytes() == b"\x15\x2b\x00\x66\x66\x66\x66\x66\xe6\x31\x40\x18"
        )

    def test_precision_float_infinity_encode(self):
        s = FloatDouble()
        s.f = float("inf")
        assert (
            s.encode().tobytes() == b"\x15\x2b\x00\x00\x00\x00\x00\x00\x00\xf0\x7f\x18"
        )

    def test_precision_float_negative_infinity_encode(self):
        s = FloatDouble()
        s.f = float("-inf")
        assert (
            s.encode().tobytes() == b"\x15\x2b\x00\x00\x00\x00\x00\x00\x00\xf0\xff\x18"
        )

    @given(v=...)
    def test_roundtrip(self, v: float):
        s = FloatDouble()
        s.f = v
        buffer = s.encode().tobytes()

        s2 = FloatDouble.decode(buffer)

        assert (
            (math.isnan(s.f) and math.isnan(s2.f))
            or (s.f > 1.8e308 and s2.f == float("inf"))
            or (s.f < -1.8e308 and s2.f == float("-inf"))
            or math.isclose(s2.f, s.f, rel_tol=2.22e-16, abs_tol=1e-15)
        )


class InnerStruct(tlv.Structure):
    a = tlv.IntMember(0, signed=True, optional=True, octets=4)
    b = tlv.IntMember(1, signed=True, optional=True, octets=4)


class OuterStruct(tlv.Structure):
    s = tlv.StructMember(0, InnerStruct)


class TestStruct:
    def test_inner_struct_decode(self):
        s = OuterStruct.decode(b"\x15\x35\x00\x20\x00\x2a\x20\x01\xef\x18\x18")
        assert_type(s, OuterStruct)
        assert_type(s.s, InnerStruct)
        assert_type(s.s.a, Optional[int])
        assert str(s) == "{\n  s = {\n    a = 42,\n    b = -17\n  }\n}"
        assert s.s.a == 42
        assert s.s.b == -17

    def test_inner_struct_decode_empty(self):
        s = OuterStruct.decode(b"\x15\x35\x00\x18\x18")
        assert str(s) == "{\n  s = {\n    \n  }\n}"
        assert s.s.a is None
        assert s.s.b is None

    def test_inner_struct_encode(self):
        s = OuterStruct()
        inner = InnerStruct()
        inner.a = 42
        inner.b = -17
        s.s = inner
        assert s.encode().tobytes() == b"\x15\x35\x00\x20\x00\x2a\x20\x01\xef\x18\x18"

    def test_inner_struct_encode_empty(self):
        s = OuterStruct()
        s.s = InnerStruct()
        assert s.encode().tobytes() == b"\x15\x35\x00\x18\x18"


class FullyQualified(tlv.Structure):
    a = tlv.IntMember((0xADA, 0xF00, 0x123), signed=True, optional=True, octets=4)
    b = tlv.IntMember((0xADA, 0xF00, 0x12345), signed=True, optional=True, octets=4)


class TestFullyQualifiedTags:
    def test_decode(self):
        s = FullyQualified.decode(
            b"\x15\xc2\xda\x0a\x00\x0f\x23\x01\x2a\x00\x00\x00\xe2\xda\x0a\x00\x0f\x45\x23\x01\x00\xef\xff\xff\xff\x18",
        )
        assert_type(s, FullyQualified)
        assert_type(s.a, Optional[int])
        assert str(s) == "{\n  a = 42,\n  b = -17\n}"
        assert s.a == 42
        assert s.b == -17

    def test_encode(self):
        s = FullyQualified()
        s.a = 42
        s.b = -17
        assert (
            s.encode().tobytes()
            == b"\x15\xc0\xda\x0a\x00\x0f\x23\x01\x2a\xe0\xda\x0a\x00\x0f\x45\x23\x01\x00\xef\x18"
        )


class InnerList(tlv.List):
    a = tlv.IntMember(0, signed=True, optional=True, octets=4)
    b = tlv.IntMember(1, signed=True, optional=True, octets=4)


class OuterStructList(tlv.Structure):
    sublist = tlv.ListMember(0, InnerList)


class TestList:
    def test_encode(self):
        s = OuterStructList()
        inner = InnerList()
        inner.a = 42
        inner.b = -17
        s.sublist = inner
        assert s.encode().tobytes() == b"\x15\x37\x00\x20\x00\x2a\x20\x01\xef\x18\x18"


class OuterStructArray(tlv.Structure):
    a = tlv.ArrayMember(0, InnerList)


class TestArray:
    def test_encode(self):
        s = OuterStructArray()
        inner = InnerList()
        inner.a = 42
        inner.b = -17
        s.a = [inner]
        assert (
            s.encode().tobytes()
            == b"\x15\x36\x00\x17\x20\x00\x2a\x20\x01\xef\x18\x18\x18"
        )

    def test_encode2(self):
        s = OuterStructArray()
        inner = InnerList()
        inner.a = 42
        inner.b = -17
        s.a = [inner, inner]
        assert (
            s.encode().tobytes()
            == b"\x15\x36\x00\x17\x20\x00\x2a\x20\x01\xef\x18\x17\x20\x00\x2a\x20\x01\xef\x18\x18\x18"
        )



import enum

from circuitmatter import data_model
from circuitmatter import tlv


class FeatureBitmap(enum.IntFlag):
    ON_OFF = 1 << 0
    LIGHTING = 1 << 1
    FREQUENCY = 1 << 2


class OptionsBitmap(data_model.Map8):
    ExecuteIfOff = 0
    CoupleColorTempToLevel = 1


class MoveModeEnum(data_model.Enum8):
    UP = 0
    DOWN = 1


class StepModeEnum(data_model.Enum8):
    UP = 0
    DOWN = 1


class LevelControl(data_model.Cluster):
    CLUSTER_ID = 0x0008

    CurrentLevel = data_model.NumberAttribute(
        0x0000, signed=False, bits=8, N_nonvolatile=True, X_nullable=True
    )
    RemainingTime = data_model.NumberAttribute(
        0x0001, signed=False, bits=16, default=0, feature=FeatureBitmap.LIGHTING
    )
    MinLevel = data_model.NumberAttribute(
        0x0002,
        signed=False,
        bits=8,
        default=lambda features: 1 if features & FeatureBitmap.LIGHTING else 0,
    )
    MaxLevel = data_model.NumberAttribute(0x0003, signed=False, bits=8, default=254)
    OnLevel = data_model.NumberAttribute(0x0011, signed=False, bits=8, X_nullable=True)
    Options = data_model.BitmapAttribute(0x000F, OptionsBitmap, default=0)
    StartUpCurrentLevel = data_model.NumberAttribute(
        0x4000,
        signed=False,
        bits=8,
        X_nullable=True,
        N_nonvolatile=True,
        feature=FeatureBitmap.LIGHTING,
    )

    class MoveToLevel(tlv.Structure):
        Level = tlv.IntMember(0, signed=False, octets=1)
        TransitionTime = tlv.IntMember(1, signed=False, octets=2, nullable=True)
        OptionsMask = tlv.BitmapMember(2, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(3, OptionsBitmap, default=0)

    move_to_level = data_model.Command(0x00, MoveToLevel)

    class Move(tlv.Structure):
        MoveMode = tlv.EnumMember(0, MoveModeEnum)
        Rate = tlv.IntMember(1, signed=True, octets=1, nullable=True)
        OptionsMask = tlv.BitmapMember(2, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(3, OptionsBitmap, default=0)

    move = data_model.Command(0x01, Move)

    class Step(tlv.Structure):
        StepMode = tlv.EnumMember(0, StepModeEnum)
        StepSize = tlv.IntMember(1, signed=True, octets=1)
        TransitionTime = tlv.IntMember(2, signed=False, octets=2, nullable=True)
        OptionsMask = tlv.BitmapMember(3, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(4, OptionsBitmap, default=0)

    step = data_model.Command(0x02, Step)

    class Stop(tlv.Structure):
        OptionsMask = tlv.BitmapMember(0, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(1, OptionsBitmap, default=0)

    stop = data_model.Command(0x03, Stop)

    move_to_level_with_on_off = data_model.Command(0x04, MoveToLevel)

    move_with_on_off = data_model.Command(0x05, Move)

    step_with_on_off = data_model.Command(0x06, Step)

    stop_with_on_off = data_model.Command(0x07, Stop)



from .on_off import OnOffLight

from circuitmatter.clusters.general import level_control


class DimmableLight(OnOffLight):
    DEVICE_TYPE_ID = 0x0101
    REVISION = 3

    def __init__(self, name):
        super().__init__(name)

        self._level_control = level_control.LevelControl()
        self._level_control.feature_map |= level_control.FeatureBitmap.LIGHTING
        self._level_control.min_level = 1
        self._level_control.max_level = 254
        self.servers.append(self._level_control)

        self._level_control.move_to_level_with_on_off = self._move_to_level_with_on_off

    def _move_to_level_with_on_off(self, session, value):
        try:
            self.brightness = value.Level / self._level_control.max_level
        except Exception as e:
            print(f"Error setting brightness: {e}")
            return
        self._level_control.CurrentLevel = value.Level

    @property
    def brightness(self):
        return self._level_control.CurrentLevel / self._level_control.max_level

    @brightness.setter
    def brightness(self, value):
        raise NotImplementedError()



from circuitmatter.data_model import (
    Enum8,
    NumberAttribute,
    ListAttribute,
    Command,
    Cluster,
)
from circuitmatter import tlv


class CertificateChainTypeEnum(Enum8):
    DAC = 1
    PAI = 2


class NodeOperationalCertStatusEnum(Enum8):
    OK = 0
    """OK, no error"""
    INVALID_PUBLIC_KEY = 1
    """Public Key in the NOC does not match the public key in the NOCSR"""
    INVALID_NODE_OP_ID = 2
    """The Node Operational ID in the NOC is not formatted correctly."""
    INVALID_NOC = 3
    """Any other validation error in NOC chain"""
    MISSING_CSR = 4
    """No record of prior CSR for which this NOC could match"""
    TABLE_FULL = 5
    """NOCs table full, cannot add another one"""
    INVALID_ADMIN_SUBJECT = 6
    """Invalid CaseAdminSubject field for an AddNOC command."""
    FABRIC_CONFLICT = 9
    """Trying to AddNOC instead of UpdateNOC against an existing Fabric."""
    LABEL_CONFLICT = 10
    """Label already exists on another Fabric."""
    INVALID_FABRIC_INDEX = 11
    """FabricIndex argument is invalid."""


RESP_MAX = 900


class NodeOperationalCredentialsCluster(Cluster):
    CLUSTER_ID = 0x003E

    class NOCStruct(tlv.Structure):
        NOC = tlv.OctetStringMember(1, 400)
        ICAC = tlv.OctetStringMember(2, 400, nullable=True)

    class FabricDescriptorStruct(tlv.Structure):
        RootPublicKey = tlv.OctetStringMember(1, 65)
        VendorID = tlv.IntMember(2, signed=False, octets=2)
        FabricID = tlv.IntMember(3, signed=False, octets=8)
        NodeID = tlv.IntMember(4, signed=False, octets=8)
        Label = tlv.UTF8StringMember(5, max_length=32, default="")

    class AttestationRequest(tlv.Structure):
        AttestationNonce = tlv.OctetStringMember(0, 32)

    class AttestationResponse(tlv.Structure):
        AttestationElements = tlv.OctetStringMember(0, RESP_MAX)
        AttestationSignature = tlv.OctetStringMember(1, 64)

    class CertificateChainRequest(tlv.Structure):
        CertificateType = tlv.EnumMember(0, CertificateChainTypeEnum)

    class CertificateChainResponse(tlv.Structure):
        Certificate = tlv.OctetStringMember(0, 600)

    class CSRRequest(tlv.Structure):
        CSRNonce = tlv.OctetStringMember(0, 32)
        IsForUpdateNOC = tlv.BoolMember(1, optional=True, default=False)

    class CSRResponse(tlv.Structure):
        NOCSRElements = tlv.OctetStringMember(0, RESP_MAX)
        AttestationSignature = tlv.OctetStringMember(1, 64)

    class AddNOC(tlv.Structure):
        NOCValue = tlv.OctetStringMember(0, 400)
        ICACValue = tlv.OctetStringMember(1, 400, optional=True)
        IPKValue = tlv.OctetStringMember(2, 16)
        CaseAdminSubject = tlv.IntMember(3, signed=False, octets=8)
        AdminVendorId = tlv.IntMember(4, signed=False, octets=2)

    class UpdateNOC(tlv.Structure):
        NOCValue = tlv.OctetStringMember(0, 400)
        ICACValue = tlv.OctetStringMember(1, 400, optional=True)

    class NOCResponse(tlv.Structure):
        StatusCode = tlv.EnumMember(0, NodeOperationalCertStatusEnum)
        FabricIndex = tlv.IntMember(1, signed=False, octets=1, optional=True)
        DebugText = tlv.UTF8StringMember(2, max_length=128, optional=True)

    class UpdateFabricLabel(tlv.Structure):
        Label = tlv.UTF8StringMember(0, max_length=32)

    class RemoveFabric(tlv.Structure):
        FabricIndex = tlv.IntMember(0, signed=False, octets=1)

    class AddTrustedRootCertificate(tlv.Structure):
        RootCACertificate = tlv.OctetStringMember(0, 400)

    nocs = ListAttribute(
        0, NOCStruct, N_nonvolatile=True, C_changes_omitted=True, default=[]
    )
    fabrics = ListAttribute(1, FabricDescriptorStruct, N_nonvolatile=True, default=[])
    supported_fabrics = NumberAttribute(2, signed=False, bits=8, F_fixed=True)
    commissioned_fabrics = NumberAttribute(
        3, signed=False, bits=8, N_nonvolatile=True, default=0
    )
    trusted_root_certificates = ListAttribute(
        4,
        tlv.OctetStringMember(None, 400),
        N_nonvolatile=True,
        C_changes_omitted=True,
        default=[],
    )
    # This attribute is weird because it is fabric sensitive but not marked as such.
    # Cluster sets current_fabric_index for use in fabric sensitive attributes and
    # happens to make this work as well.
    current_fabric_index = NumberAttribute(5, signed=False, bits=8, default=0)

    attestation_request = Command(0x00, AttestationRequest, 0x01, AttestationResponse)

    certificate_chain_request = Command(
        0x02, CertificateChainRequest, 0x03, CertificateChainResponse
    )

    csr_request = Command(0x04, CSRRequest, 0x05, CSRResponse)

    add_noc = Command(0x06, AddNOC, 0x08, NOCResponse)

    update_noc = Command(0x07, UpdateNOC, 0x08, NOCResponse)

    update_fabric_label = Command(0x09, UpdateFabricLabel, 0x08, NOCResponse)

    remove_fabric = Command(0x0A, RemoveFabric, 0x08, NOCResponse)

    add_trusted_root_certificate = Command(0x0B, AddTrustedRootCertificate, None, None)



# This file should only be needed when generating certificates.

import binascii
import hashlib

from . import tlv
from . import pase
from .data_model import Enum8

import ecdsa
from ecdsa.curves import NIST256p
from ecdsa import der

PAI_KEY_DER = b"\x30\x77\x02\x01\x01\x04\x20\xbb\x76\xa5\x80\x5f\x97\x26\x49\xaf\x1e\x8a\x87\xdc\x45\x57\xe6\x2c\x09\x00\xe5\x07\x09\xe8\x5c\x79\xc6\x44\xdf\x78\x90\xe5\x96\xa0\x0a\x06\x08\x2a\x86\x48\xce\x3d\x03\x01\x07\xa1\x44\x03\x42\x00\x04\x37\x5d\x2b\xc8\xc6\x15\x27\x5b\xfd\x84\x8b\x52\xfe\x21\x96\xe2\xa1\x4e\xf3\xcc\x91\xae\xf0\x5d\xff\x85\x1c\xbc\x19\xb1\xa9\x35\x45\x8c\xfe\x04\xaa\x42\x4e\x01\x6d\xe3\xd6\x74\xdc\x5b\x73\x29\xbd\x77\x57\xfd\xdb\x32\x38\xd6\x26\x73\x62\x9b\x3c\x79\x08\x45"

INVALID_PASSCODES = [
    0,
    11111111,
    22222222,
    33333333,
    44444444,
    55555555,
    66666666,
    77777777,
    88888888,
    12345678,
    87654321,
]


class CertificationType(Enum8):
    DEVELOPMENT_AND_TEST = 0
    PROVISIONAL = 1
    OFFICIAL = 2


class CertificationDeclaration(tlv.Structure):
    format_version = tlv.IntMember(0, signed=False, octets=1)
    vendor_id = tlv.IntMember(1, signed=False, octets=2)
    product_id_array = tlv.ArrayMember(
        2, tlv.IntMember(0, signed=False, octets=2), max_length=100
    )
    device_type_id = tlv.IntMember(3, signed=False, octets=4)
    certificate_id = tlv.UTF8StringMember(4, max_length=19)
    security_level = tlv.IntMember(5, signed=False, octets=1)
    security_information = tlv.IntMember(6, signed=False, octets=2)
    version_number = tlv.IntMember(7, signed=False, octets=2)
    certification_type = tlv.EnumMember(8, CertificationType)
    dac_origin_vendor_id = tlv.IntMember(9, signed=False, octets=2, optional=True)
    dac_origin_product_id = tlv.IntMember(10, signed=False, octets=2, optional=True)
    authorized_paa_list = tlv.ArrayMember(
        11, tlv.OctetStringMember(None, max_length=20), optional=True, max_length=10
    )


def encode_set(*encoded_pieces):
    total_len = sum([len(p) for p in encoded_pieces])
    return b"\x31" + der.encode_length(total_len) + b"".join(encoded_pieces)


def encode_utf8_string(s):
    encoded = s.encode("utf-8")
    return b"\x0c" + der.encode_length(len(encoded)) + encoded


def generate_certificates(
    vendor_id=0xFFF1, product_id=0x8000, device_type=22, prefix=None
):
    declaration = CertificationDeclaration()
    declaration.format_version = 1  # Always 1
    declaration.vendor_id = vendor_id
    declaration.product_id_array = [product_id]
    declaration.device_type_id = device_type
    declaration.certificate_id = "CSA00000SWC00000-00"
    declaration.security_level = 0  # Always 0
    declaration.security_information = 0  # Always 0
    declaration.version_number = 1  # Always 1
    declaration.certification_type = CertificationType.DEVELOPMENT_AND_TEST
    declaration = declaration.encode()

    # From: https://github.com/project-chip/matter.js/blob/main/packages/protocol/src/certificate/CertificationDeclarationManager.ts
    # NIST256p is the same as secp256r1
    private_key = ecdsa.keys.SigningKey.from_string(
        b"\xae\xf3\x48\x41\x16\xe9\x48\x1e\xc5\x7b\xe0\x47\x2d\xf4\x1b\xf4\x99\x06\x4e\x50\x24\xad\x86\x9e\xca\x5e\x88\x98\x02\xd4\x80\x75",
        curve=ecdsa.curves.NIST256p,
        hashfunc=hashlib.sha256,
    )
    subject_key_identifier = b"\x62\xfa\x82\x33\x59\xac\xfa\xa9\x96\x3e\x1c\xfa\x14\x0a\xdd\xf5\x04\xf3\x71\x60"
    signature = private_key.sign_deterministic(
        declaration,
        hashfunc=hashlib.sha256,
        sigencode=ecdsa.util.sigencode_der_canonize,
    )

    certification_declaration = []
    # version
    certification_declaration.append(der.encode_integer(3))
    # Digest algorithm
    certification_declaration.append(
        encode_set(der.encode_sequence(der.encode_oid(2, 16, 840, 1, 101, 3, 4, 2, 1)))
    )
    # encap content info
    encap_content_info = []
    # content type
    encap_content_info.append(der.encode_oid(1, 2, 840, 113549, 1, 7, 1))
    # content
    encap_content_info.append(
        der.encode_constructed(0, der.encode_octet_string(declaration))
    )
    certification_declaration.append(der.encode_sequence(*encap_content_info))

    signer_info = []
    # version
    signer_info.append(der.encode_integer(3))
    # subject key identifier
    signer_info.append(
        b"\x80"
        + der.encode_length(len(subject_key_identifier))
        + subject_key_identifier
    )
    # digest algorithm
    signer_info.append(
        der.encode_sequence(der.encode_oid(2, 16, 840, 1, 101, 3, 4, 2, 1))
    )
    # signature algorithm
    signer_info.append(der.encode_sequence(der.encode_oid(1, 2, 840, 10045, 4, 3, 2)))
    # signature
    signer_info.append(der.encode_octet_string(signature))
    certification_declaration.append(encode_set(der.encode_sequence(*signer_info)))

    signed_data = []
    signed_data.append(der.encode_oid(1, 2, 840, 113549, 1, 7, 2))
    cd = der.encode_sequence(*certification_declaration)
    signed_data.append(der.encode_constructed(0, cd))
    cms_signed = der.encode_sequence(*signed_data)

    return cms_signed


def generate_dac(
    vendor_id, product_id, product_name, random_source
) -> tuple[bytes, bytes]:
    dac_key = ecdsa.keys.SigningKey.generate(
        curve=ecdsa.NIST256p, hashfunc=hashlib.sha256, entropy=random_source.urandom
    )

    version = der.encode_constructed(0, der.encode_integer(2))
    serial_number = der.encode_integer(1)
    signature_algorithm = der.encode_sequence(der.encode_oid(1, 2, 840, 10045, 4, 3, 2))
    # CircuitMatter PAI for vendor ID 0xfff4
    issuer = b"\x30\x32\x31\x1a\x30\x18\x06\x03\x55\x04\x03\x0c\x11\x43\x69\x72\x63\x75\x69\x74\x4d\x61\x74\x74\x65\x72\x20\x50\x41\x49\x31\x14\x30\x12\x06\x0a\x2b\x06\x01\x04\x01\x82\xa2\x7c\x02\x01\x0c\x04\x46\x46\x46\x34"

    # Starting 10/17/2024 and never expiring
    validity = b"\x30\x20\x17\x0d\x32\x34\x31\x30\x31\x37\x30\x30\x30\x30\x30\x30\x5a\x18\x0f\x39\x39\x39\x39\x31\x32\x33\x31\x32\x33\x35\x39\x35\x39\x5a"

    common_name = encode_set(
        der.encode_sequence(
            der.encode_oid(2, 5, 4, 3), encode_utf8_string(product_name)
        )
    )
    encoded_vendor_id = encode_set(
        der.encode_sequence(
            der.encode_oid(1, 3, 6, 1, 4, 1, 37244, 2, 1),
            encode_utf8_string(f"{vendor_id:04X}"),
        )
    )
    encoded_product_id = encode_set(
        der.encode_sequence(
            der.encode_oid(1, 3, 6, 1, 4, 1, 37244, 2, 2),
            encode_utf8_string(f"{product_id:04X}"),
        )
    )
    subject = der.encode_sequence(common_name, encoded_vendor_id, encoded_product_id)

    algorithm_id = b"\x30\x13\x06\x07\x2a\x86\x48\xce\x3d\x02\x01\x06\x08\x2a\x86\x48\xce\x3d\x03\x01\x07"

    public_key = dac_key.verifying_key.to_string(encoding="uncompressed")
    public_key_info = der.encode_sequence(
        algorithm_id, der.encode_bitstring(public_key, unused=0)
    )

    basic_constraints = b"\x30\x0c\x06\x03\x55\x1d\x13\x01\x01\xff\x04\x02\x30\x00"
    key_usage = b"\x30\x0e\x06\x03\x55\x1d\x0f\x01\x01\xff\x04\x04\x03\x02\x07\x80"
    key_id = der.encode_sequence(
        der.encode_oid(2, 5, 29, 14),
        der.encode_octet_string(
            der.encode_octet_string(hashlib.sha1(public_key).digest())
        ),
    )
    # ID of the CircuitMatter 0xFFF4 PAI
    authority_key_id = b"\x30\x1f\x06\x03\x55\x1d\x23\x04\x18\x30\x16\x80\x14\x07\xf8\x38\x0a\x5f\x01\x36\xfc\xe2\x36\xbd\x45\xf2\x88\xff\x22\xdc\xa6\xf4\xa7"
    extensions = der.encode_constructed(
        3, der.encode_sequence(basic_constraints, key_usage, key_id, authority_key_id)
    )

    certificate = der.encode_sequence(
        version,
        serial_number,
        signature_algorithm,
        issuer,
        validity,
        subject,
        public_key_info,
        extensions,
    )

    pai_key = ecdsa.keys.SigningKey.from_der(PAI_KEY_DER, hashfunc=hashlib.sha256)
    signature = pai_key.sign_deterministic(
        certificate,
        hashfunc=hashlib.sha256,
        sigencode=ecdsa.util.sigencode_der_canonize,
    )

    dac_cert = der.encode_sequence(
        certificate, signature_algorithm, der.encode_bitstring(signature, unused=0)
    )
    dac_key = dac_key.to_der()
    return dac_cert, dac_key


def compute_verifier(passcode: int, salt: bytes, iterations: int) -> bytes:
    w0, w1 = pase._pbkdf2(passcode, salt, iterations)
    L = NIST256p.generator * w1

    return w0.to_bytes(NIST256p.baselen, byteorder="big") + L.to_bytes("uncompressed")


# Look up tables for Verhoeff Algorithm
# From: https://en.wikipedia.org/wiki/Verhoeff_algorithm#Table-based_algorithm
D_TABLE = (
    b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09"
    + b"\x01\x02\x03\x04\x00\x06\x07\x08\x09\x05"
    + b"\x02\x03\x04\x00\x01\x07\x08\x09\x05\x06"
    + b"\x03\x04\x00\x01\x02\x08\x09\x05\x06\x07"
    + b"\x04\x00\x01\x02\x03\x09\x05\x06\x07\x08"
    + b"\x05\x09\x08\x07\x06\x00\x04\x03\x02\x01"
    + b"\x06\x05\x09\x08\x07\x01\x00\x04\x03\x02"
    + b"\x07\x06\x05\x09\x08\x02\x01\x00\x04\x03"
    + b"\x08\x07\x06\x05\x09\x03\x02\x01\x00\x04"
    + b"\x09\x08\x07\x06\x05\x04\x03\x02\x01\x00"
)

INV_TABLE = b"\x00\x04\x03\x02\x01\x05\x06\x07\x08\x09"

P_TABLE = (
    b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09"
    + b"\x01\x05\x07\x06\x02\x08\x03\x00\x09\x04"
    + b"\x05\x08\x00\x03\x07\x09\x06\x01\x04\x02"
    b"\x08\x09\x01\x06\x00\x04\x03\x05\x02\x07"
    + b"\x09\x04\x05\x03\x01\x02\x06\x08\x07\x00"
    + b"\x04\x02\x08\x06\x05\x07\x03\x09\x00\x01"
    + b"\x02\x07\x09\x03\x08\x00\x06\x04\x01\x05"
    + b"\x07\x00\x04\x06\x09\x01\x03\x02\x05\x08"
)


def _bcd(buf, n):
    div = 10 ** (len(buf) - 1)
    for i in range(len(buf)):
        buf[i] = (n // div) % 10
        div //= 10


def compute_manual_code(
    discriminator, passcode, vendor_id=None, product_id=None
) -> str:
    vid_pid_present = 0
    if vendor_id is not None and product_id is not None:
        vid_pid_present = 1

    digits = memoryview(bytearray(11))
    digits[0] = (vid_pid_present << 2) | (discriminator >> 10)
    d2_6 = ((discriminator & 0x300) << 6) | (passcode & 0x3FFF)
    _bcd(digits[1:6], d2_6)
    d7_10 = passcode >> 14
    _bcd(digits[6:10], d7_10)

    # Checksum of zero. We'll overwrite it.
    digits[10] = 0

    c = 0
    for i, n in enumerate(reversed(digits)):
        c = D_TABLE[c * 10 + P_TABLE[(i % 8) * 10 + n]]

    digits[10] = INV_TABLE[c]
    digits = [str(x) for x in digits]
    digits.insert(4, "-")
    digits.insert(8, "-")

    return "".join(digits)


def generate_initial_state(vendor_id, product_id, product_name, random_source):
    if vendor_id != 0xFFF4 or product_id != 0x1234:
        raise ValueError("Invalid vendor_id or product_id")

    cd = generate_certificates(vendor_id=vendor_id, product_id=product_id)

    dac_cert, dac_key = generate_dac(vendor_id, product_id, product_name, random_source)

    passcode = 0
    while passcode in INVALID_PASSCODES:
        passcode = random_source.randbelow(99999999)
    discriminator = random_source.randbelow(1 << 12)  # A 12-bit random number
    iteration_count = 10000
    salt = random_source.urandom(32)
    verifier = compute_verifier(passcode, salt, iteration_count)
    # This does *NOT* follow the spec because the passcode is stored alongside the verifier.
    # The spec wants the passcode stored physically on the box and package of the device in
    # the setup code and QR Code. The verifier is in the device only.
    initial_state = {
        "discriminator": discriminator,
        "passcode": passcode,
        "manual_code": compute_manual_code(discriminator, passcode),
        "iteration-count": iteration_count,
        "salt": binascii.b2a_base64(salt, newline=False).decode("utf-8"),
        "verifier": binascii.b2a_base64(verifier, newline=False).decode("utf-8"),
        "devices": {
            "root": {
                "0x3e": {
                    "cd": binascii.b2a_base64(cd, newline=False).decode("utf-8"),
                    "dac_cert": binascii.b2a_base64(dac_cert, newline=False).decode(
                        "utf-8"
                    ),
                    "dac_key": binascii.b2a_base64(dac_key, newline=False).decode(
                        "utf-8"
                    ),
                }
            },
        },
    }
    return initial_state


if __name__ == "__main__":
    from circuitmatter.utility import random

    initial_state = generate_initial_state(0xFFF4, 0x1234, "CircuitMatter", random)
    print(initial_state)



"""Pure Python implementation of the Matter IOT protocol."""

import json
import pathlib
import socket
import time

import circuitmatter as cm

from circuitmatter.device_types.lighting import extended_color

from circuitmatter.utility import random
from circuitmatter.utility.recording import RecordingSocketPool, RecordingRandom
from circuitmatter.utility.replay import ReplaySocketPool, ReplayRandom

from circuitmatter.utility.mdns import DummyMDNS
from circuitmatter.utility.mdns.avahi import Avahi


class NeoPixel(extended_color.ExtendedColorLight):
    @property
    def color_rgb(self):
        return self._color

    @color_rgb.setter
    def color_rgb(self, value):
        self._color = value
        print(f"new color 0x{value:06x}")

    @property
    def brightness(self):
        return self._brightness

    @brightness.setter
    def brightness(self, value):
        self._brightness = value
        print(f"new brightness {value}")

    def on(self):
        print("on!")

    def off(self):
        print("off!")


def run(replay_file=None):
    device_state = pathlib.Path("live-rgb-device-state.json")
    if replay_file:
        replay_lines = []
        with open(replay_file, "r") as f:
            device_state_fn = f.readline().strip()
            for line in f:
                replay_lines.append(json.loads(line))
        socketpool = ReplaySocketPool(replay_lines)
        mdns_server = DummyMDNS()
        random_source = ReplayRandom(replay_lines)
        # Reset device state to before the captured run
        if device_state_fn == "none":
            device_state.unlink(missing_ok=True)
        else:
            device_state.write_text(pathlib.Path(device_state_fn).read_text())
    else:
        timestamp = time.strftime("%Y%m%d-%H%M%S")
        record_file = open(f"test_data/recorded_packets-{timestamp}.jsonl", "w")
        device_state_fn = f"test_data/device_state-{timestamp}.json"
        replay_device_state = pathlib.Path(device_state_fn)
        if device_state.exists():
            record_file.write(f"{device_state_fn}\n")
            # Save device state before we run so replays can use it.
            replay_device_state.write_text(device_state.read_text())
        else:
            # No starting state.
            record_file.write("none\n")
        socketpool = RecordingSocketPool(record_file, socket)
        mdns_server = Avahi()
        random_source = RecordingRandom(record_file, random)

    matter = cm.CircuitMatter(socketpool, mdns_server, random_source, device_state)
    led = NeoPixel("neopixel1")
    matter.add_device(led)
    while True:
        matter.process_packets()


if __name__ == "__main__":
    import sys

    replay_file = None
    if len(sys.argv) > 1:
        replay_file = sys.argv[1]
    run(replay_file=replay_file)



from . import tlv

import ecdsa
import enum
import hashlib
import hmac
import struct

# Section 3.6

SYMMETRIC_KEY_LENGTH_BITS = 128
SYMMETRIC_KEY_LENGTH_BYTES = 16
AEAD_MIC_LENGTH_BITS = 128
AEAD_MIC_LENGTH_BYTES = 16
AEAD_NONCE_LENGTH_BYTES = 13

GROUP_SIZE_BITS = 256
GROUP_SIZE_BYTES = 32
PUBLIC_KEY_SIZE_BYTES = (2 * GROUP_SIZE_BYTES) + 1

HASH_LEN_BITS = 256
HASH_LEN_BYTES = 32
HASH_BLOCK_LEN_BYTES = 64

# Upper limit for encoded certificate size.
CERTIFICATE_SIZE = 400


class DNAttribute(tlv.List):
    # Section 6.5.6.1

    common_name = tlv.UTF8StringMember(1, 100)
    surname = tlv.UTF8StringMember(2, 100)
    serial_num = tlv.UTF8StringMember(3, 100)
    country_name = tlv.UTF8StringMember(4, 100)
    locality_name = tlv.UTF8StringMember(5, 100)
    state_or_province_name = tlv.UTF8StringMember(6, 100)
    org_name = tlv.UTF8StringMember(7, 100)
    org_unit_name = tlv.UTF8StringMember(8, 100)
    title = tlv.UTF8StringMember(9, 100)
    name = tlv.UTF8StringMember(10, 100)
    given_name = tlv.UTF8StringMember(11, 100)
    initials = tlv.UTF8StringMember(12, 100)
    gen_qualifier = tlv.UTF8StringMember(13, 100)
    dn_qualifier = tlv.UTF8StringMember(14, 100)
    pseudonym = tlv.UTF8StringMember(15, 100)
    domain_component = tlv.OctetStringMember(16, 100)
    matter_node_id = tlv.IntMember(17, signed=False, octets=8)
    matter_firmware_signing_id = tlv.IntMember(18, signed=False, octets=8)
    matter_icac_id = tlv.IntMember(19, signed=False, octets=8)
    matter_rcac_id = tlv.IntMember(20, signed=False, octets=8)
    matter_fabric_id = tlv.IntMember(21, signed=False, octets=8)
    matter_noc_cat = tlv.IntMember(22, signed=False, octets=8)
    common_name_ps = tlv.OctetStringMember(129, 100)
    surname_ps = tlv.OctetStringMember(130, 100)
    serial_num_ps = tlv.OctetStringMember(131, 100)
    country_name_ps = tlv.OctetStringMember(132, 100)
    locality_name_ps = tlv.OctetStringMember(133, 100)
    state_or_province_name_ps = tlv.OctetStringMember(134, 100)
    org_name_ps = tlv.OctetStringMember(135, 100)
    org_unit_name_ps = tlv.OctetStringMember(136, 100)
    title_ps = tlv.OctetStringMember(137, 100)
    name_ps = tlv.OctetStringMember(138, 100)
    given_name_ps = tlv.OctetStringMember(139, 100)
    initials_ps = tlv.OctetStringMember(140, 100)
    gen_qualifier_ps = tlv.OctetStringMember(141, 100)
    dn_qualifier_ps = tlv.OctetStringMember(142, 100)
    pseudonym_ps = tlv.OctetStringMember(143, 100)


class BasicContraints(tlv.Structure):
    # Section 6.5.11.1
    is_ca = tlv.BoolMember(1)
    path_len_constraint = tlv.IntMember(2, signed=False, octets=1, optional=True)


class Extensions(tlv.List):
    # Section 6.5.11
    basic_cnstr = tlv.StructMember(1, BasicContraints)
    key_usage = tlv.IntMember(2, signed=False, octets=2)
    extended_key_usage = tlv.ArrayMember(
        3, tlv.IntMember(None, signed=False, octets=1), max_length=100
    )
    subject_key_id = tlv.OctetStringMember(4, 20)
    authority_key_id = tlv.OctetStringMember(5, 20)
    future_extension = tlv.OctetStringMember(6, 400)


class SignatureAlgorithm(enum.IntEnum):
    # Section 6.5.5
    ECDSA_WITH_SHA256 = 1


class PublicKeyAlgorithm(enum.IntEnum):
    # Section 6.5.8
    EC_PUB_KEY = 1


class EllipticCurveId(enum.IntEnum):
    # Section 6.5.9
    PRIME256V1 = 1


class MatterCertificate(tlv.Structure):
    # Section 6.5.2

    serial_num = tlv.OctetStringMember(1, 20)
    sig_algo = tlv.EnumMember(2, SignatureAlgorithm)
    issuer = tlv.ListMember(3, DNAttribute)
    not_before = tlv.IntMember(4, signed=False, octets=4)
    not_after = tlv.IntMember(5, signed=False, octets=4)
    subject = tlv.ListMember(6, DNAttribute)
    pub_key_algo = tlv.EnumMember(7, PublicKeyAlgorithm)
    ec_curve_id = tlv.EnumMember(8, EllipticCurveId)
    ec_pub_key = tlv.OctetStringMember(9, 65)
    extensions = tlv.ListMember(10, Extensions)
    signature = tlv.OctetStringMember(11, GROUP_SIZE_BYTES * 2)


def Hash(*message) -> bytes:
    h = hashlib.sha256()
    for m in message:
        h.update(m)
    return h.digest()


def HMAC(key, message) -> bytes:
    m = hmac.new(key, digestmod=hashlib.sha256)
    m.update(message)
    return m.digest()


def HKDF_Extract(salt, input_key) -> bytes:
    return HMAC(salt, input_key)


def HKDF_Expand(prk, info, length) -> bytes:
    if length > 255:
        raise ValueError("length must be less than 256")
    last_hash = b""
    bytes_generated = []
    num_bytes_generated = 0
    i = 1
    while num_bytes_generated < length:
        num_bytes_generated += HASH_LEN_BYTES
        # Do the hmac directly so we don't need to allocate a buffer for last_hash + info + i.
        m = hmac.new(prk, digestmod=hashlib.sha256)
        m.update(last_hash)
        m.update(info)
        m.update(struct.pack("b", i))
        last_hash = m.digest()
        bytes_generated.append(last_hash)
        i += 1
    return b"".join(bytes_generated)


def KDF(input_key, salt, info, length):
    if salt is None:
        salt = b"\x00" * HASH_LEN_BYTES
    return HKDF_Expand(HKDF_Extract(salt, input_key), info, length // 8)[: length // 8]


def ECDH(private_key: ecdsa.keys.SigningKey, public_key: bytes) -> bytes:
    ecdh = ecdsa.ecdh.ECDH(ecdsa.NIST256p)
    ecdh.load_private_key(private_key)
    ecdh.load_received_public_key_bytes(public_key)
    return ecdh.generate_sharedsecret_bytes()



from circuitmatter.clusters.general.identify import Identify
from circuitmatter.clusters.general import on_off

from .. import simple_device


class OnOffLight(simple_device.SimpleDevice):
    DEVICE_TYPE_ID = 0x0100
    REVISION = 3

    def __init__(self, name):
        super().__init__(name)

        self._identify = Identify()
        self.servers.append(self._identify)

        self._on_off = on_off.OnOff()
        self._on_off.on = self._on
        self._on_off.off = self._off
        self._on_off.feature_map |= on_off.FeatureBitmap.LIGHTING
        self.servers.append(self._on_off)

    def _on(self, session):
        try:
            self.on()
        except Exception as e:
            print(f"Error turning on light: {e}")
            return
        self._on_off.on_off = True

    def _off(self, session):
        try:
            self.off()
        except Exception as e:
            print(f"Error turning off light: {e}")
            return
        self._on_off.on_off = False

    def on(self):
        raise NotImplementedError()

    def off(self):
        raise NotImplementedError()



from . import crypto
from . import protocol
from . import session
from . import tlv


class CASEMessage(tlv.Structure):
    PROTOCOL_ID = protocol.ProtocolId.SECURE_CHANNEL


class Sigma1(CASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.CASE_SIGMA1

    initiatorRandom = tlv.OctetStringMember(1, 32)
    initiatorSessionId = tlv.IntMember(2, signed=False, octets=2)
    destinationId = tlv.OctetStringMember(3, crypto.HASH_LEN_BYTES)
    initiatorEphPubKey = tlv.OctetStringMember(4, crypto.PUBLIC_KEY_SIZE_BYTES)
    initiatorSessionParams = tlv.StructMember(
        5, session.SessionParameterStruct, optional=True
    )
    resumptionID = tlv.OctetStringMember(6, 16, optional=True)
    initiatorResumeMIC = tlv.OctetStringMember(
        7, crypto.AEAD_MIC_LENGTH_BYTES, optional=True
    )


class Sigma2TbsData(tlv.Structure):
    responderNOC = tlv.OctetStringMember(1, crypto.CERTIFICATE_SIZE)
    responderICAC = tlv.OctetStringMember(2, crypto.CERTIFICATE_SIZE, optional=True)
    responderEphPubKey = tlv.OctetStringMember(3, crypto.PUBLIC_KEY_SIZE_BYTES)
    initiatorEphPubKey = tlv.OctetStringMember(4, crypto.PUBLIC_KEY_SIZE_BYTES)


class Sigma2TbeData(tlv.Structure):
    responderNOC = tlv.OctetStringMember(1, crypto.CERTIFICATE_SIZE)
    responderICAC = tlv.OctetStringMember(2, crypto.CERTIFICATE_SIZE, optional=True)
    signature = tlv.OctetStringMember(3, crypto.GROUP_SIZE_BYTES * 2)
    resumptionID = tlv.OctetStringMember(4, 16)


class Sigma2(CASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.CASE_SIGMA2
    responderRandom = tlv.OctetStringMember(1, 32)
    responderSessionId = tlv.IntMember(2, signed=False, octets=2)
    responderEphPubKey = tlv.OctetStringMember(3, crypto.PUBLIC_KEY_SIZE_BYTES)
    encrypted2 = tlv.OctetStringMember(4, Sigma2TbeData.max_length())
    responderSessionParams = tlv.StructMember(
        5, session.SessionParameterStruct, optional=True
    )


class Sigma3TbsData(tlv.Structure):
    initiatorNOC = tlv.OctetStringMember(1, crypto.CERTIFICATE_SIZE)
    initiatorICAC = tlv.OctetStringMember(2, crypto.CERTIFICATE_SIZE, optional=True)
    initiatorEphPubKey = tlv.OctetStringMember(3, crypto.PUBLIC_KEY_SIZE_BYTES)
    responderEphPubKey = tlv.OctetStringMember(4, crypto.PUBLIC_KEY_SIZE_BYTES)


class Sigma3TbeData(tlv.Structure):
    initiatorNOC = tlv.OctetStringMember(1, crypto.CERTIFICATE_SIZE)
    initiatorICAC = tlv.OctetStringMember(2, crypto.CERTIFICATE_SIZE, optional=True)
    signature = tlv.OctetStringMember(3, crypto.GROUP_SIZE_BYTES * 2)


class Sigma3(CASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.CASE_SIGMA3
    encrypted3 = tlv.OctetStringMember(1, Sigma3TbeData.max_length())


class Sigma2Resume(CASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.CASE_SIGMA2_RESUME
    resumptionID = tlv.OctetStringMember(1, 16)
    sigma2ResumeMIC = tlv.OctetStringMember(2, 16)
    responderSessionID = tlv.IntMember(3, signed=False, octets=2)
    responderSessionParams = tlv.StructMember(
        4, session.SessionParameterStruct, optional=True
    )


def compute_destination_id(
    root_public_key, fabric_id, node_id, initiator_random, identity_protection_key
):
    destination_message = b"".join(
        (initiator_random, root_public_key, fabric_id, node_id)
    )
    return crypto.HMAC(identity_protection_key, destination_message)


if __name__ == "__main__":
    root_public_key = bytes.fromhex(
        "04:4a:9f:42:b1:ca:48:40:d3:72:92:bb:c7:f6:a7:e1:1e:22:20:0c:97:6f:c9:00:db:c9:8a:7a:38:3a:64:1c:b8:25:4a:2e:56:d4:e2:95:a8:47:94:3b:4e:38:97:c4:a7:73:e9:30:27:7b:4d:9f:be:de:8a:05:26:86:bf:ac:fa".replace(
            ":", ""
        )
    )
    fabric_id = bytes.fromhex("62:d3:15:d1:08:c9:06:29".replace(":", ""))
    node_id = bytes.fromhex("14:ef:13:7b:aa:44:55:cd".replace(":", ""))
    initiator_random = bytes.fromhex(
        "7e:17:12:31:56:8d:fa:17:20:6b:3a:cc:f8:fa:ec:2f:4d:21:b5:80:11:31:96:f4:7c:7c:4d:eb:81:0a:73:dc".replace(
            ":", ""
        )
    )
    identity_protection_key = bytes.fromhex(
        " 9b:c6:1c:d9:c6:2a:2d:f6:d6:4d:fc:aa:9d:c4:72:d4".replace(":", "")
    )

    destination_id = compute_destination_id(
        root_public_key, fabric_id, node_id, initiator_random, identity_protection_key
    )
    print(destination_id.hex(":"))



import enum

from circuitmatter import data_model
from circuitmatter import tlv


class FeatureBitmap(enum.IntFlag):
    LIGHTING = 1 << 0
    DEAD_FRONT_BEHAVIOR = 1 << 1
    OFF_ONLY = 1 << 2


class StartUpOnOffEnum(data_model.Enum8):
    OFF = 0
    ON = 1
    TOGGLE = 2


class OnOff(data_model.Cluster):
    CLUSTER_ID = 0x0006

    OnOff = data_model.BoolAttribute(0x0000, default=False, N_nonvolatile=True)
    GlobalSceneControl = data_model.BoolAttribute(
        0x4000, default=True, feature=FeatureBitmap.LIGHTING
    )
    OnTime = data_model.NumberAttribute(
        0x4001, signed=False, bits=16, default=0, feature=FeatureBitmap.LIGHTING
    )
    OffWaitTime = data_model.NumberAttribute(
        0x4002, signed=False, bits=16, default=0, feature=FeatureBitmap.LIGHTING
    )
    StartUpOnOff = data_model.EnumAttribute(
        0x4003,
        StartUpOnOffEnum,
        N_nonvolatile=True,
        X_nullable=True,
        feature=FeatureBitmap.LIGHTING,
    )

    off = data_model.Command(0x00, None)
    on = data_model.Command(0x01, None)
    toggle = data_model.Command(0x02, None)

    class OffWithEffect(tlv.Structure):
        EffectIdentifier = tlv.EnumMember(0, 0)
        EffectVariant = tlv.EnumMember(1, 0, default=0)

    off_with_effect = data_model.Command(0x40, OffWithEffect)
    on_with_recall_global_scene = data_model.Command(0x41, None)
    on_with_timed_off = data_model.Command(0x42, None)



from circuitmatter.data_model import (
    Cluster,
    ClusterId,
    DeviceTypeId,
    EndpointNumber,
    ListAttribute,
    Uint16,
)
from circuitmatter import tlv


class DescriptorCluster(Cluster):
    CLUSTER_ID = 0x001D

    class DeviceTypeStruct(tlv.Structure):
        DeviceType = DeviceTypeId(0)
        Revision = Uint16(1, minimum=1)

    DeviceTypeList = ListAttribute(0x0000, DeviceTypeStruct)
    ServerList = ListAttribute(0x0001, ClusterId())
    ClientList = ListAttribute(0x0002, ClusterId())
    PartsList = ListAttribute(0x0003, EndpointNumber())



import binascii
import enum
import inspect
import random
import struct
import traceback
import typing
from typing import Iterable, Union

from . import interaction_model
from . import tlv

ATTRIBUTES_KEY = "a"


class Enum8(enum.IntEnum):
    pass


class Enum16(enum.IntEnum):
    pass


class Map8(enum.IntFlag):
    pass


class Map16(enum.IntFlag):
    pass


class Uint16(tlv.IntMember):
    def __init__(self, _id=None, minimum=0, **kwargs):
        super().__init__(_id, signed=False, octets=2, minimum=minimum, **kwargs)


class Uint32(tlv.IntMember):
    def __init__(self, _id=None, minimum=0, **kwargs):
        super().__init__(_id, signed=False, octets=4, minimum=minimum, **kwargs)


class Uint64(tlv.IntMember):
    def __init__(self, _id=None, minimum=0, **kwargs):
        super().__init__(_id, signed=False, octets=8, minimum=minimum, **kwargs)


class NodeId(Uint64):
    pass


class GroupId(Uint16):
    def __init__(self, _id=None, **kwargs):
        super().__init__(_id, minimum=1, **kwargs)


class ClusterId(Uint16):
    pass


class AttributeId(Uint16):
    pass


class EventId(Uint16):
    pass


class CommandId(Uint16):
    pass


class DeviceTypeId(Uint32):
    pass


class EndpointNumber(Uint16):
    def __init__(self, _id=None, **kwargs):
        super().__init__(_id, minimum=1, **kwargs)


# Data model "lists" are encoded as tlv arrays. 🙄
class List(tlv.ArrayMember):
    pass


class Attribute:
    def __init__(
        self,
        _id,
        default=None,
        optional=False,
        feature=0,
        C_changes_omitted=False,
        F_fixed=False,
        N_nonvolatile=False,
        P_reportable=False,
        Q_quieter_reporting=False,
        S_scene=False,
        X_nullable=False,
    ):
        self.id = _id
        self.default = default
        self.optional = optional
        self.feature = feature
        self.nullable = X_nullable
        self.nonvolatile = N_nonvolatile

    def __get__(self, instance, cls):
        v = instance._attribute_values.get(self.id, None)
        if v is None:
            if callable(self.default):
                return self.default(instance.feature_map)
            return self.default
        return v

    def __set__(self, instance, value):
        old_value = instance._attribute_values.get(self.id, None)
        if old_value == value:
            return
        instance._attribute_values[self.id] = value
        if self.nonvolatile:
            instance._nonvolatile[ATTRIBUTES_KEY][hex(self.id)] = self.to_json(value)
        instance.data_version += 1

    def to_json(self, value):
        return value

    def from_json(self, value):
        return value

    def encode(self, value) -> bytes:
        if value is None and self.nullable:
            return b"\x14"  # No tag, NULL
        return self._encode(value)

    def _encode(self, value):
        raise NotImplementedError()


class NumberAttribute(Attribute):
    def __init__(self, _id, *, signed, bits, **kwargs):
        self.signed = signed
        self.bits = bits
        self.id = _id
        super().__init__(_id, **kwargs)

    @staticmethod
    def encode_number(value, *, signed=True) -> bytes:
        bit_length = value.bit_length()
        format_string = None
        if signed:
            type = tlv.ElementType.SIGNED_INT
        else:
            type = tlv.ElementType.UNSIGNED_INT
        length = 0  # in power of two
        if bit_length <= 8:
            format_string = "<Bb" if signed else "<BB"
            length = 0
        elif bit_length <= 16:
            format_string = "<Bh" if signed else "<BH"
            length = 1
        elif bit_length <= 32:
            format_string = "<Bi" if signed else "<BI"
            length = 2
        else:
            format_string = "<Bq" if signed else "<BQ"
            length = 3

        return struct.pack(format_string, type | length, value)

    def _encode(self, value) -> bytes:
        return NumberAttribute.encode_number(value, signed=self.signed)


class EnumAttribute(NumberAttribute):
    def __init__(self, _id, enum_type, **kwargs):
        self.enum_type = enum_type
        bits = 8 if issubclass(enum_type, Enum8) else 16
        super().__init__(_id, signed=False, bits=bits, **kwargs)


class _PersistentList:
    def __init__(self, wrapped_list, attribute, instance):
        self._list = wrapped_list
        self._instance = instance
        self._attribute = attribute

    def append(self, value):
        self._list.append(value)
        self._instance._nonvolatile[ATTRIBUTES_KEY][hex(self._attribute.id)] = (
            self._attribute.to_json(self._list)
        )

    def __getitem__(self, index):
        return self._list[index]

    def __setitem__(self, index, value):
        self._list[index] = value
        self._dirty = True

    def __iter__(self):
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def __str__(self):
        return "persistent" + str(self._list)


class ListAttribute(Attribute):
    def __init__(self, _id, element_type, **kwargs):
        if inspect.isclass(element_type) and issubclass(element_type, enum.Enum):
            element_type = tlv.EnumMember(None, element_type)
        self.tlv_type = tlv.ArrayMember(None, element_type)
        self._element_type = element_type
        # Copy the default list so we don't accidentally share it with another
        # cluster of the same type.
        if "default" in kwargs and isinstance(kwargs["default"], list):
            kwargs["default"] = list(kwargs["default"])
        super().__init__(_id, **kwargs)

    def __get__(self, instance, cls):
        v = super().__get__(instance, cls)
        if self.nonvolatile and v is not None and not isinstance(v, _PersistentList):
            # Wrap the list in an object that tracks changes and writes them to nonvolatile.
            p = _PersistentList(v, self, instance)
            instance._attribute_values[self.id] = p
            return p
        return v

    def to_json(self, value):
        return [
            binascii.b2a_base64(self._element_type.encode(v), newline=False).decode(
                "utf-8"
            )
            for v in value
        ]

    def from_json(self, value):
        return [
            self._element_type.decode(memoryview(binascii.a2b_base64(v))) for v in value
        ]

    def _encode(self, value) -> bytes:
        return self.tlv_type.encode(value)

    def element_from_value(self, value):
        if issubclass(self._element_type, tlv.Container):
            return self._element_type.from_value(value)
        if issubclass(self._element_type, enum.Enum):
            return self._element_type(value)
        return value


class BoolAttribute(Attribute):
    def encode(self, value) -> bytes:
        return struct.pack("B", tlv.ElementType.BOOL | (1 if value else 0))


class StructAttribute(Attribute):
    def __init__(self, _id, struct_type, default=None):
        self.struct_type = struct_type
        super().__init__(_id, default=default)

    def encode(self, value) -> memoryview:
        buffer = memoryview(bytearray(value.max_length() + 2))
        buffer[0] = tlv.ElementType.STRUCTURE
        end = value.encode_into(buffer, 1)
        return buffer[:end]


class OctetStringAttribute(Attribute):
    def __init__(self, _id, min_length, max_length, **kwargs):
        self.min_length = min_length
        self.max_length = max_length
        self.member = tlv.OctetStringMember(None, max_length=max_length)
        super().__init__(_id, **kwargs)

    def encode(self, value):
        return self.member.encode(value)


class UTF8StringAttribute(Attribute):
    def __init__(self, _id, min_length=0, max_length=1200, **kwargs):
        self.min_length = min_length
        self.max_length = max_length
        self.member = tlv.UTF8StringMember(None, max_length=max_length)
        super().__init__(_id, **kwargs)

    def encode(self, value):
        return self.member.encode(value)


class BitmapAttribute(NumberAttribute):
    def __init__(self, _id, enum_type, **kwargs):
        self.enum_type = enum_type
        bits = 8 if issubclass(enum_type, Map8) else 16
        super().__init__(_id, signed=False, bits=bits, **kwargs)


class Command:
    def __init__(
        self,
        command_id,
        request_type,
        response_id=None,
        response_type=interaction_model.StatusCode,
    ):
        self.command_id = command_id
        self.request_type = request_type
        self.response_id = response_id
        self.response_type = response_type


class Cluster:
    cluster_revision = NumberAttribute(0xFFFD, signed=False, bits=12, default=1)
    feature_map = NumberAttribute(0xFFFC, signed=False, bits=32, default=0)
    attribute_list = ListAttribute(0xFFFB, AttributeId())
    event_list = ListAttribute(0xFFFA, EventId())
    accepted_command_list = ListAttribute(0xFFF9, CommandId())
    generated_command_list = ListAttribute(0xFFF8, CommandId())

    def __init__(self):
        self._attribute_values = {}
        # Use random since this isn't for security or replayability.
        self.data_version = random.randint(0, 0xFFFFFFFF)

        self.attribute_list = []
        for _, descriptor in self._attributes():
            self.attribute_list.append(descriptor.id)

        self.accepted_command_list = []
        self.generated_command_list = []
        self.event_list = []
        for _, descriptor in self._commands():
            self.accepted_command_list.append(descriptor.command_id)
            if descriptor.response_id is not None:
                self.generated_command_list.append(descriptor.response_id)

    def __contains__(self, descriptor_id):
        return descriptor_id in self._attribute_values

    @classmethod
    def _attributes(cls) -> Iterable[tuple[str, Attribute]]:
        for superclass in cls.__mro__:
            for field_name, descriptor in vars(superclass).items():
                if not field_name.startswith("_") and isinstance(descriptor, Attribute):
                    yield field_name, descriptor

    def restore(self, nonvolatile):
        self._nonvolatile = nonvolatile

        if ATTRIBUTES_KEY not in nonvolatile:
            nonvolatile[ATTRIBUTES_KEY] = {}
        for field_name, descriptor in self._attributes():
            if descriptor.nonvolatile:
                if hex(descriptor.id) in nonvolatile[ATTRIBUTES_KEY]:
                    # Update our live value
                    self._attribute_values[descriptor.id] = descriptor.from_json(
                        nonvolatile[ATTRIBUTES_KEY][hex(descriptor.id)]
                    )
                else:
                    # Store the default
                    nonvolatile[ATTRIBUTES_KEY][hex(descriptor.id)] = descriptor.default

    def get_attribute_data(
        self, session, path
    ) -> typing.List[interaction_model.AttributeDataIB]:
        replies = []
        for field_name, descriptor in self._attributes():
            if path.Attribute is not None and descriptor.id != path.Attribute:
                continue
            if descriptor.feature and not (self.feature_map & descriptor.feature):
                continue
            self.current_fabric_index = session.local_fabric_index
            value = getattr(self, field_name)
            self.current_fabric_index = None
            print(
                "reading",
                f"EP{path.Endpoint}",
                type(self).__name__,
                field_name,
                "->",
                value,
            )
            if value is None and descriptor.optional:
                continue
            data = interaction_model.AttributeDataIB()
            data.DataVersion = 0
            attribute_path = interaction_model.AttributePathIB()
            attribute_path.Endpoint = path.Endpoint
            attribute_path.Cluster = path.Cluster
            attribute_path.Attribute = descriptor.id
            data.Path = attribute_path
            data.Data = descriptor.encode(value)
            print(
                f"{path.Endpoint}/{path.Cluster:x}/{descriptor.id:x} -> {data.Data.hex(' ')}"
            )
            replies.append(data)
            if path.Attribute is not None:
                break
        if not replies:
            print(f"\033[91mnot found 0x{path.Attribute:04x}\033[0m")
        return replies

    def set_attribute(
        self, context, attribute_data
    ) -> interaction_model.AttributeStatusIB:
        status_code = interaction_model.StatusCode.SUCCESS
        for field_name, descriptor in self._attributes():
            path = attribute_data.Path
            if path.Attribute is not None and descriptor.id != path.Attribute:
                continue
            has_list_index = False
            for entry in path:
                if (
                    isinstance(entry, tuple)
                    and entry[0] == interaction_model.AttributePathIB.ListIndex
                ):
                    has_list_index = True
                    break
            # value =
            value = attribute_data.Data
            print("writing", self, field_name, "->", value, "?", has_list_index)

            if has_list_index:
                if not isinstance(descriptor, ListAttribute):
                    status_code = interaction_model.StatusCode.UNSUPPORTED_WRITE
                    break
                list_ = getattr(self, field_name)
                if not isinstance(list_, list):
                    status_code = interaction_model.StatusCode.UNSUPPORTED_WRITE
                    break
                list_.append(descriptor.element_from_value(value))
            else:
                setattr(self, field_name, value)
        astatus = interaction_model.AttributeStatusIB()
        astatus.Path = attribute_data.Path
        status = interaction_model.StatusIB()
        status.Status = status_code
        status.ClusterStatus = 0
        astatus.Status = status
        return astatus

    @classmethod
    def _commands(cls) -> Iterable[tuple[str, Command]]:
        for superclass in cls.__mro__:
            for field_name, descriptor in vars(superclass).items():
                if not field_name.startswith("_") and isinstance(descriptor, Command):
                    yield field_name, descriptor

    def invoke(
        self, session, path, fields
    ) -> Union[interaction_model.CommandDataIB, interaction_model.StatusCode, None]:
        found = False
        for field_name, descriptor in self._commands():
            if descriptor.command_id != path.Command:
                continue

            print("invoke", type(self).__name__, field_name)
            command = getattr(self, field_name)
            if callable(command):
                if descriptor.request_type is not None:
                    try:
                        arg = descriptor.request_type.from_value(fields)
                    except ValueError:
                        return interaction_model.StatusCode.INVALID_COMMAND
                    try:
                        print(arg)
                        result = command(session, arg)
                    except Exception as e:
                        traceback.print_exception(e)
                        return interaction_model.StatusCode.FAILURE
                else:
                    try:
                        result = command(session)
                    except Exception as e:
                        print(e)
                        return interaction_model.StatusCode.FAILURE
            else:
                return interaction_model.StatusCode.UNSUPPORTED_COMMAND
            if descriptor.response_type is interaction_model.StatusCode:
                if result is None:
                    return interaction_model.StatusCode.SUCCESS
                return result
            elif descriptor.response_type is not None:
                cdata = interaction_model.CommandDataIB()
                response_path = interaction_model.CommandPathIB()
                response_path.Endpoint = path.Endpoint
                response_path.Cluster = path.Cluster
                response_path.Command = descriptor.response_id
                cdata.CommandPath = response_path
                if result:
                    cdata.CommandFields = descriptor.response_type.encode(result)
                return cdata
            return result
        if not found:
            print("not found", path.Command)
        return None



class DummyMDNS:
    def advertise_service(
        self,
        service_type,
        protocol,
        port,
        txt_records=[],
        subtypes=[],
        instance_name="",
    ):
        print(f"Advertise service {service_type} {protocol} {port} {txt_records}")



from circuitmatter.data_model import (
    Enum8,
    NumberAttribute,
    StructAttribute,
    EnumAttribute,
    BoolAttribute,
    Command,
    Cluster,
)
from circuitmatter import tlv


class CommissioningErrorEnum(Enum8):
    OK = 0
    VALUE_OUTSIDE_RANGE = 1
    INVALID_AUTHENTICATION = 2
    NO_FAIL_SAFE = 3
    BUSY_WITH_OTHER_ADMIN = 4


class RegulatoryLocationType(Enum8):
    INDOOR = 0
    OUTDOOR = 1
    INDOOR_OUTDOOR = 2


class GeneralCommissioningCluster(Cluster):
    CLUSTER_ID = 0x0030

    class BasicCommissioningInfo(tlv.Structure):
        FailSafeExpiryLengthSeconds = tlv.IntMember(0, signed=False, octets=2)
        MaxCumulativeFailsafeSeconds = tlv.IntMember(1, signed=False, octets=2)

    breadcrumb = NumberAttribute(0, signed=False, bits=64, default=0)
    basic_commissioning_info = StructAttribute(1, BasicCommissioningInfo)
    regulatory_config = EnumAttribute(
        2, RegulatoryLocationType, default=RegulatoryLocationType.INDOOR_OUTDOOR
    )
    location_capability = EnumAttribute(
        3, RegulatoryLocationType, default=RegulatoryLocationType.INDOOR_OUTDOOR
    )
    support_concurrent_connection = BoolAttribute(4, default=True)

    class ArmFailSafe(tlv.Structure):
        ExpiryLengthSeconds = tlv.IntMember(0, signed=False, octets=2, default=900)
        Breadcrumb = tlv.IntMember(1, signed=False, octets=8)

    class CommissioningResponse(tlv.Structure):
        ErrorCode = tlv.EnumMember(
            0, CommissioningErrorEnum, default=CommissioningErrorEnum.OK
        )
        DebugText = tlv.UTF8StringMember(1, max_length=128, default="")

    ArmFailSafeResponse = CommissioningResponse

    arm_fail_safe = Command(0x00, ArmFailSafe, 0x01, ArmFailSafeResponse)

    class SetRegulatoryConfig(tlv.Structure):
        NewRegulatoryConfig = tlv.EnumMember(0, RegulatoryLocationType)
        CountryCode = tlv.UTF8StringMember(1, max_length=2)
        Breadcrumb = tlv.IntMember(2, signed=False, octets=8)

    SetRegulatoryConfigResponse = CommissioningResponse

    set_regulatory_config = Command(
        0x02, SetRegulatoryConfig, 0x03, SetRegulatoryConfigResponse
    )

    CommissioningCompleteResponse = CommissioningResponse

    commissioning_complete = Command(0x04, None, 0x05, CommissioningCompleteResponse)



import enum

from circuitmatter import data_model, tlv


class FeatureBitmap(enum.IntFlag):
    HUE_SATURATION = 1 << 0
    ENHANCED_HUE = 1 << 1
    COLOR_LOOP = 1 << 2
    XY = 1 << 3
    COLOR_TEMPERATURE = 1 << 4


class OptionsBitmap(data_model.Map8):
    EXECUTE_IF_OFF = 1 << 0


class Direction(data_model.Enum8):
    SHORTEST_DISTANCE = 0
    LONGEST_DISTANCE = 1
    UP = 2
    DOWN = 3


class MoveMode(data_model.Enum8):
    STOP = 0
    UP = 1
    DOWN = 3


class StepMode(data_model.Enum8):
    UP = 1
    DOWN = 3


class ColorMode(data_model.Enum8):
    HUE_SATURATION = 0
    XY = 1
    COLOR_TEMPERATURE = 2


class ColorControl(data_model.Cluster):
    CLUSTER_ID = 0x0300
    cluster_revision = 6

    current_hue = data_model.NumberAttribute(0x0000, signed=False, bits=8, default=0)
    current_saturation = data_model.NumberAttribute(
        0x0001, signed=False, bits=8, default=0
    )
    remaining_time = data_model.NumberAttribute(
        0x0002, signed=False, bits=16, default=0
    )
    current_x = data_model.NumberAttribute(0x0003, signed=False, bits=16, default=0)
    current_y = data_model.NumberAttribute(0x0004, signed=False, bits=16, default=0)
    drift_compensation = data_model.NumberAttribute(
        0x0005, signed=False, bits=8, default=0
    )
    compensation_text = data_model.UTF8StringAttribute(0x0006, default="")
    color_temperature = data_model.NumberAttribute(
        0x0007, signed=False, bits=16, default=0
    )
    color_mode = data_model.EnumAttribute(0x0008, data_model.Enum8, default=0)
    options = data_model.BitmapAttribute(0x000F, OptionsBitmap, default=0)
    color_capabilities = data_model.BitmapAttribute(0x400A, FeatureBitmap, default=0)

    color_temp_physical_min_mireds = data_model.NumberAttribute(
        0x400B,
        signed=False,
        bits=16,
        default=0,
        feature=FeatureBitmap.COLOR_TEMPERATURE,
    )  # maximum=0xfeff
    color_temp_physical_max_mireds = data_model.NumberAttribute(
        0x400C,
        signed=False,
        bits=16,
        default=0xFEFF,
        feature=FeatureBitmap.COLOR_TEMPERATURE,
    )  # maximum=0xfeff

    class MoveToHue(tlv.Structure):
        Hue = tlv.IntMember(0, signed=False, octets=1, maximum=254)
        Direction = tlv.EnumMember(1, Direction)
        TransitionTime = tlv.IntMember(2, signed=False, octets=2, nullable=True)
        OptionsMask = tlv.BitmapMember(3, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(4, OptionsBitmap, default=0)

    move_to_hue = data_model.Command(0x00, MoveToHue)

    class MoveHue(tlv.Structure):
        MoveMode = tlv.EnumMember(0, Direction)
        Rate = tlv.IntMember(1, signed=True, octets=1, nullable=True)
        OptionsMask = tlv.BitmapMember(2, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(3, OptionsBitmap, default=0)

    move_hue = data_model.Command(0x01, MoveHue)

    class StepHue(tlv.Structure):
        StepMode = tlv.EnumMember(0, StepMode)
        StepSize = tlv.IntMember(1, signed=True, octets=1)
        TransitionTime = tlv.IntMember(2, signed=False, octets=1)
        OptionsMask = tlv.BitmapMember(3, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(4, OptionsBitmap, default=0)

    step_hue = data_model.Command(0x02, StepHue)

    class MoveToSaturation(tlv.Structure):
        Saturation = tlv.IntMember(0, signed=False, octets=1)
        TransitionTime = tlv.IntMember(1, signed=False, octets=2)
        OptionsMask = tlv.BitmapMember(2, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(3, OptionsBitmap, default=0)

    move_to_saturation = data_model.Command(0x03, MoveToSaturation)

    class MoveSaturation(tlv.Structure):
        MoveMode = tlv.EnumMember(0, MoveMode)
        Rate = tlv.IntMember(1, signed=True, octets=1, nullable=True)
        OptionsMask = tlv.BitmapMember(2, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(3, OptionsBitmap, default=0)

    move_saturation = data_model.Command(0x04, MoveSaturation)

    class StepSaturation(tlv.Structure):
        StepMode = tlv.EnumMember(0, StepMode)
        StepSize = tlv.IntMember(1, signed=True, octets=1)
        TransitionTime = tlv.IntMember(2, signed=False, octets=2, nullable=True)
        OptionsMask = tlv.BitmapMember(3, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(4, OptionsBitmap, default=0)

    step_saturation = data_model.Command(0x05, StepSaturation)

    class MoveToHueAndSaturation(tlv.Structure):
        Hue = tlv.IntMember(0, signed=False, octets=1)
        Saturation = tlv.IntMember(1, signed=False, octets=1)
        TransitionTime = tlv.IntMember(2, signed=False, octets=2, nullable=True)
        OptionsMask = tlv.BitmapMember(3, OptionsBitmap, default=0)
        OptionsOverride = tlv.BitmapMember(4, OptionsBitmap, default=0)

    move_to_hue_and_saturation = data_model.Command(0x06, MoveToHueAndSaturation)



import binascii
import json
import time


class RecordingRandom:
    def __init__(self, record_file, random):
        self.record_file = record_file
        self._random = random

    def urandom(self, nbytes):
        data = self._random.urandom(nbytes)
        entry = (
            "urandom",
            time.monotonic_ns(),
            nbytes,
            binascii.b2a_base64(data, newline=False).decode("utf-8"),
        )
        json.dump(entry, self.record_file)
        self.record_file.write("\n")
        return data

    def randbelow(self, n):
        value = self._random.randbelow(n)
        entry = ("randbelow", time.monotonic_ns(), n, value)
        json.dump(entry, self.record_file)
        self.record_file.write("\n")
        return value


class RecordingSocket:
    def __init__(self, record_file, socket):
        self.record_file = record_file
        self.socket = socket

    def bind(self, address):
        self.socket.bind(address)

    def setblocking(self, value):
        self.socket.setblocking(value)

    def recvfrom_into(self, buffer, nbytes=None):
        nbytes, addr = self.socket.recvfrom_into(buffer, nbytes)
        entry = (
            "receive",
            time.monotonic_ns(),
            addr,
            binascii.b2a_base64(buffer[:nbytes], newline=False).decode("utf-8"),
        )
        json.dump(entry, self.record_file)
        self.record_file.write("\n")
        return nbytes, addr

    def sendto(self, data, address):
        entry = (
            "send",
            time.monotonic_ns(),
            address,
            binascii.b2a_base64(data, newline=False).decode("utf-8"),
        )
        json.dump(entry, self.record_file)
        self.record_file.write("\n")
        return self.socket.sendto(data, address)


class RecordingSocketPool:
    def __init__(self, record_file, socket):
        self.AF_INET6 = socket.AF_INET6
        self.SOCK_DGRAM = socket.SOCK_DGRAM
        self.record_file = record_file
        self._socket_created = False
        self._socket = socket

    def socket(self, *args, **kwargs):
        if self._socket_created:
            raise RuntimeError("Only one socket can be created")
        self._socket_created = True
        return RecordingSocket(self.record_file, self._socket.socket(*args, **kwargs))



"""Pure Python implementation of the Matter IOT protocol."""

import binascii
import hashlib
import json
import pathlib
import time

from . import case
from . import interaction_model
from . import nonvolatile
from .message import Message
from .protocol import InteractionModelOpcode, ProtocolId, SecureProtocolOpcode
from . import session
from .device_types.utility.root_node import RootNode

__version__ = "0.2.3"


class CircuitMatter:
    def __init__(
        self,
        socketpool=None,
        mdns_server=None,
        random_source=None,
        state_filename="matter-device-state.json",
        vendor_id=0xFFF4,
        product_id=0x1234,
        product_name="CircuitMatter Device",
    ):
        if socketpool is None:
            import socket

            socketpool = socket
        self.socketpool = socketpool

        if mdns_server is None:
            from circuitmatter.utility.mdns.avahi import Avahi

            mdns_server = Avahi()
        self.mdns_server = mdns_server

        if random_source is None:
            from circuitmatter.utility import random

            random_source = random
        self.random = random_source

        state_file = pathlib.Path(state_filename)
        if not state_file.exists():
            from circuitmatter import certificates

            initial_state = certificates.generate_initial_state(
                vendor_id, product_id, product_name, random_source
            )
            with open(state_filename, "w") as f:
                json.dump(initial_state, f, indent=1)

        self.nonvolatile = nonvolatile.PersistentDictionary(state_filename)

        for key in ["discriminator", "salt", "iteration-count", "verifier"]:
            if key not in self.nonvolatile:
                raise RuntimeError(f"Missing key {key} in state file")

        self.packet_buffer = memoryview(bytearray(1280))

        # Define the UDP IP address and port
        UDP_IP = "::"  # Listen on all available network interfaces
        self.UDP_PORT = 5541

        # Create the UDP socket
        self.socket = self.socketpool.socket(
            self.socketpool.AF_INET6, self.socketpool.SOCK_DGRAM
        )

        # Bind the socket to the IP and port
        self.socket.bind((UDP_IP, self.UDP_PORT))
        print(f"Listening on UDP port {self.UDP_PORT}")
        self.socket.setblocking(False)

        self._endpoints = {}
        self._next_endpoint = 0
        self.root_node = RootNode(
            random_source,
            self.mdns_server,
            self.UDP_PORT,
            vendor_id,
            product_id,
            version=__version__,
        )
        self.add_device(self.root_node)

        self.vendor_id = vendor_id
        self.product_id = product_id
        self.manager = session.SessionManager(
            self.random, self.socket, self.root_node.noc
        )

        if self.root_node.fabric_count == 0:
            self.start_commissioning()

    def start_commissioning(self):
        discriminator = self.nonvolatile["discriminator"]
        passcode = self.nonvolatile["passcode"]
        txt_records = {
            "PI": "",
            "PH": "33",
            "CM": "1",
            "D": str(discriminator),
            "CRI": "3000",
            "CRA": "4000",
            "T": "1",
            "VP": f"{self.vendor_id}+{self.product_id}",
        }
        from . import pase

        pase.show_qr_code(self.vendor_id, self.product_id, discriminator, passcode)
        print("Manual code:", self.nonvolatile["manual_code"])
        instance_name = self.random.urandom(8).hex().upper()
        self.mdns_server.advertise_service(
            "_matterc",
            "_udp",
            self.UDP_PORT,
            txt_records=txt_records,
            instance_name=instance_name,
            subtypes=[
                f"_L{discriminator}._sub._matterc._udp",
                "_CM._sub._matterc._udp",
            ],
        )

    def add_cluster(self, endpoint, cluster):
        if endpoint not in self._endpoints:
            self._endpoints[endpoint] = {}
            if endpoint > 0:
                self.root_node.descriptor.PartsList.append(endpoint)
            self._next_endpoint = max(self._next_endpoint, endpoint + 1)
        self._endpoints[endpoint][cluster.CLUSTER_ID] = cluster

    def add_device(self, device):
        self._endpoints[self._next_endpoint] = {}
        if self._next_endpoint > 0:
            self.root_node.descriptor.PartsList.append(self._next_endpoint)

        for server in device.servers:
            device.descriptor.ServerList.append(server.CLUSTER_ID)
            self.add_cluster(self._next_endpoint, server)
        self.add_cluster(self._next_endpoint, device.descriptor)

        if "devices" not in self.nonvolatile:
            self.nonvolatile["devices"] = {}
        if device.name not in self.nonvolatile["devices"]:
            self.nonvolatile["devices"][device.name] = {}
        device.restore(self.nonvolatile["devices"][device.name])
        self._next_endpoint += 1

    def process_packets(self):
        while True:
            try:
                nbytes, addr = self.socket.recvfrom_into(
                    self.packet_buffer, len(self.packet_buffer)
                )
            except BlockingIOError:
                break
            if nbytes == 0:
                break

            self.process_packet(addr, self.packet_buffer[:nbytes])

    def _build_attribute_error(self, path, status_code):
        report = interaction_model.AttributeReportIB()
        astatus = interaction_model.AttributeStatusIB()
        astatus.Path = path
        status = interaction_model.StatusIB()
        status.Status = status_code
        status.ClusterStatus = 0
        astatus.Status = status
        report.AttributeStatus = astatus
        return report

    def get_report(self, context, cluster, path):
        reports = []
        datas = cluster.get_attribute_data(context, path)
        for data in datas:
            report = interaction_model.AttributeReportIB()
            report.AttributeData = data
            reports.append(report)
        # Only add status if an error occurs
        if not datas:
            print("Unsupported attribute", cluster, path)
            report = self._build_attribute_error(
                path, interaction_model.StatusCode.UNSUPPORTED_ATTRIBUTE
            )
            reports.append(report)
        return reports

    def invoke(self, session, cluster, path, fields, command_ref):
        print("invoke", path)
        response = interaction_model.InvokeResponseIB()
        cdata = cluster.invoke(session, path, fields)
        if isinstance(cdata, interaction_model.CommandDataIB):
            if command_ref is not None:
                cdata.CommandRef = command_ref
            response.Command = cdata
        else:
            cstatus = interaction_model.CommandStatusIB()
            cstatus.CommandPath = path
            status = interaction_model.StatusIB()
            if cdata is None:
                status.Status = interaction_model.StatusCode.UNSUPPORTED_COMMAND
                print("UNSUPPORTED_COMMAND")
            else:
                status.Status = cdata
            cstatus.Status = status
            if command_ref is not None:
                cstatus.CommandRef = command_ref
            response.Status = cstatus
            return response

        return response

    def read_attribute_path(self, context, path):
        attribute_reports = []
        if path.Endpoint is None:
            endpoints = self._endpoints
        else:
            endpoints = [path.Endpoint]

        # Wildcard so we get it from every endpoint.
        for endpoint in endpoints:
            temp_path = path.copy()
            temp_path.Endpoint = endpoint
            if path.Cluster is None:
                clusters = self._endpoints[endpoint].values()
            else:
                if path.Cluster not in self._endpoints[endpoint]:
                    print(
                        f"Cluster 0x{path.Cluster:02x} not found on endpoint {endpoint}"
                    )
                    continue
                clusters = [self._endpoints[endpoint][path.Cluster]]
            for cluster in clusters:
                temp_path.Cluster = cluster.CLUSTER_ID
                attribute_reports.extend(self.get_report(context, cluster, temp_path))
        return attribute_reports

    def process_packet(self, address, data):
        # Print the received data and the address of the sender
        # This is section 4.7.2
        message = Message()
        message.decode(data)
        message.source_ipaddress = address
        if message.secure_session:
            secure_session_context = None
            if message.session_id < len(self.manager.secure_session_contexts):
                secure_session_context = self.manager.secure_session_contexts[
                    message.session_id
                ]
            if secure_session_context is None:
                print("Failed to find session. Ignoring.")
                return
            # Decrypt the payload
            if not secure_session_context.decrypt_and_verify(message):
                print("Failed to decrypt message. Ignoring.")
                return
        message.parse_protocol_header()
        self.manager.mark_duplicate(message)

        exchange = self.manager.process_exchange(message)
        if exchange is None:
            print(f"Dropping message {message.message_counter}")
            return
        else:
            print(
                f"Processing message {message.message_counter} for exchange {exchange.exchange_id}"
            )

        protocol_id = message.protocol_id
        protocol_opcode = message.protocol_opcode

        if protocol_id == ProtocolId.SECURE_CHANNEL:
            if protocol_opcode == SecureProtocolOpcode.MSG_COUNTER_SYNC_REQ:
                print("Received Message Counter Synchronization Request")
            elif protocol_opcode == SecureProtocolOpcode.MSG_COUNTER_SYNC_RSP:
                print("Received Message Counter Synchronization Response")
            elif protocol_opcode == SecureProtocolOpcode.PBKDF_PARAM_REQUEST:
                print("Received PBKDF Parameter Request")
                from . import pase

                # This is Section 4.14.1.2
                request = pase.PBKDFParamRequest.decode(message.application_payload)
                exchange.commissioning_hash = hashlib.sha256(
                    b"CHIP PAKE V1 Commissioning"
                )
                exchange.commissioning_hash.update(message.application_payload)
                if request.passcodeId == 0:
                    pass
                    # Send back failure
                    # response = StatusReport()
                    # response.GeneralCode
                # print(request)
                response = pase.PBKDFParamResponse()
                response.initiatorRandom = request.initiatorRandom

                # Generate a random number
                response.responderRandom = self.random.urandom(32)
                session_context = self.manager.new_context()
                response.responderSessionId = session_context.local_session_id
                exchange.secure_session_context = session_context
                session_context.peer_session_id = request.initiatorSessionId
                if not request.hasPBKDFParameters:
                    params = pase.Crypto_PBKDFParameterSet()
                    params.iterations = self.nonvolatile["iteration-count"]
                    params.salt = binascii.a2b_base64(self.nonvolatile["salt"])
                    response.pbkdf_parameters = params

                encoded = response.encode()
                exchange.commissioning_hash.update(encoded)
                exchange.send(response)

            elif protocol_opcode == SecureProtocolOpcode.PBKDF_PARAM_RESPONSE:
                print("Received PBKDF Parameter Response")
            elif protocol_opcode == SecureProtocolOpcode.PASE_PAKE1:
                from . import pase

                print("Received PASE PAKE1")
                pake1 = pase.PAKE1.decode(message.application_payload)
                pake2 = pase.PAKE2()
                verifier = binascii.a2b_base64(self.nonvolatile["verifier"])
                context = exchange.commissioning_hash.digest()
                del exchange.commissioning_hash

                cA, Ke = pase.compute_verification(
                    self.random, pake1, pake2, context, verifier
                )
                exchange.cA = cA
                exchange.Ke = Ke
                exchange.send(pake2)
            elif protocol_opcode == SecureProtocolOpcode.PASE_PAKE2:
                print("Received PASE PAKE2")
                raise NotImplementedError("Implement SPAKE2+ prover")
            elif protocol_opcode == SecureProtocolOpcode.PASE_PAKE3:
                from . import pase

                print("Received PASE PAKE3")
                pake3 = pase.PAKE3.decode(message.application_payload)
                if pake3.cA != exchange.cA:
                    del exchange.cA
                    del exchange.Ke
                    print("cA mismatch")
                    error_status = session.StatusReport()
                    error_status.general_code = session.GeneralCode.FAILURE
                    error_status.protocol_id = ProtocolId.SECURE_CHANNEL
                    error_status.protocol_code = (
                        session.SecureChannelProtocolCode.INVALID_PARAMETER
                    )
                    exchange.send(error_status)
                else:
                    exchange.session.session_timestamp = time.monotonic()
                    status_ok = session.StatusReport()
                    status_ok.general_code = session.GeneralCode.SUCCESS
                    status_ok.protocol_id = ProtocolId.SECURE_CHANNEL
                    status_ok.protocol_code = (
                        session.SecureChannelProtocolCode.SESSION_ESTABLISHMENT_SUCCESS
                    )
                    exchange.send(status_ok)

                    # Fully initialize the secure session context we'll use going
                    # forwards.
                    secure_session_context = exchange.secure_session_context

                    # Compute session keys
                    pase.compute_session_keys(exchange.Ke, secure_session_context)
                    print("PASE succeeded")
            elif protocol_opcode == SecureProtocolOpcode.CASE_SIGMA1:
                print("Received CASE Sigma1")
                sigma1 = case.Sigma1.decode(message.application_payload)
                response = self.manager.reply_to_sigma1(exchange, sigma1)

                exchange.send(response)
            elif protocol_opcode == SecureProtocolOpcode.CASE_SIGMA2:
                print("Received CASE Sigma2")
            elif protocol_opcode == SecureProtocolOpcode.CASE_SIGMA3:
                print("Received CASE Sigma3")
                sigma3 = case.Sigma3.decode(message.application_payload)
                protocol_code = self.manager.reply_to_sigma3(exchange, sigma3)

                error_status = session.StatusReport()
                general_code = session.GeneralCode.FAILURE
                if (
                    protocol_code
                    == session.SecureChannelProtocolCode.SESSION_ESTABLISHMENT_SUCCESS
                ):
                    general_code = session.GeneralCode.SUCCESS
                error_status.general_code = general_code
                error_status.protocol_id = ProtocolId.SECURE_CHANNEL
                error_status.protocol_code = protocol_code
                exchange.send(error_status)
            elif protocol_opcode == SecureProtocolOpcode.CASE_SIGMA2_RESUME:
                print("Received CASE Sigma2 Resume")
            elif protocol_opcode == SecureProtocolOpcode.STATUS_REPORT:
                print("Received Status Report")
                report = session.StatusReport()
                report.decode(message.application_payload)
                print(report)

                # Acknowledge the message because we have no further reply.
                if message.exchange_flags & session.ExchangeFlags.R:
                    exchange.send_standalone()
            elif protocol_opcode == SecureProtocolOpcode.ICD_CHECK_IN:
                print("Received ICD Check-in")
            elif protocol_opcode == SecureProtocolOpcode.MRP_STANDALONE_ACK:
                print("Received MRP Standalone Ack")
                print(message)
            else:
                print("Unhandled secure channel opcode", protocol_opcode)
        elif message.protocol_id == ProtocolId.INTERACTION_MODEL:
            secure_session_context = self.manager.secure_session_contexts[
                message.session_id
            ]
            if protocol_opcode == InteractionModelOpcode.READ_REQUEST:
                read_request = interaction_model.ReadRequestMessage.decode(
                    message.application_payload
                )
                attribute_reports = []
                for path in read_request.AttributeRequests:
                    print("read", path)
                    attribute_reports.extend(
                        self.read_attribute_path(secure_session_context, path)
                    )
                response = interaction_model.ReportDataMessage()
                response.AttributeReports = attribute_reports
                exchange.send(response)
            elif protocol_opcode == InteractionModelOpcode.WRITE_REQUEST:
                print("Received Write Request")
                write_request = interaction_model.WriteRequestMessage.decode(
                    message.application_payload
                )
                write_responses = []
                for request in write_request.WriteRequests:
                    path = request.Path
                    if path.Cluster in self._endpoints[path.Endpoint]:
                        cluster = self._endpoints[path.Endpoint][path.Cluster]
                        write_responses.append(
                            cluster.set_attribute(secure_session_context, request)
                        )
                response = interaction_model.WriteResponseMessage()
                response.WriteResponses = write_responses
                exchange.send(response)

            elif protocol_opcode == InteractionModelOpcode.INVOKE_REQUEST:
                print("Received Invoke Request")
                invoke_request = interaction_model.InvokeRequestMessage.decode(
                    message.application_payload
                )
                for invoke in invoke_request.InvokeRequests:
                    path = invoke.CommandPath
                    invoke_responses = []
                    if path.Endpoint is None:
                        # Wildcard so we get it from every endpoint.
                        for endpoint in self._endpoints:
                            if path.Cluster in self._endpoints[endpoint]:
                                cluster = self._endpoints[endpoint][path.Cluster]
                                path.Endpoint = endpoint
                                invoke_responses.append(
                                    self.invoke(
                                        secure_session_context,
                                        cluster,
                                        path,
                                        invoke.CommandFields,
                                    )
                                )
                            else:
                                print(f"Cluster 0x{path.Cluster:02x} not found")
                    else:
                        if path.Cluster in self._endpoints[path.Endpoint]:
                            cluster = self._endpoints[path.Endpoint][path.Cluster]
                            invoke_responses.append(
                                self.invoke(
                                    secure_session_context,
                                    cluster,
                                    path,
                                    invoke.CommandFields,
                                    invoke.CommandRef,
                                )
                            )
                        else:
                            print(f"Cluster 0x{path.Cluster:02x} not found")
                response = interaction_model.InvokeResponseMessage()
                response.SuppressResponse = False
                response.InvokeResponses = invoke_responses
                exchange.send(response)
            elif protocol_opcode == InteractionModelOpcode.INVOKE_RESPONSE:
                print("Received Invoke Response")
            elif protocol_opcode == InteractionModelOpcode.SUBSCRIBE_REQUEST:
                print("Received Subscribe Request")
                subscribe_request = interaction_model.SubscribeRequestMessage.decode(
                    message.application_payload
                )
                print(subscribe_request)
                attribute_reports = []
                for path in subscribe_request.AttributeRequests:
                    attribute_reports.extend(
                        self.read_attribute_path(secure_session_context, path)
                    )
                response = interaction_model.ReportDataMessage()
                response.SubscriptionId = exchange.exchange_id
                response.AttributeReports = attribute_reports
                exchange.send(response)
                final_response = interaction_model.SubscribeResponseMessage()
                final_response.SubscriptionId = exchange.exchange_id
                final_response.MaxInterval = subscribe_request.MaxIntervalCeiling
                exchange.queue(final_response)
            elif protocol_opcode == InteractionModelOpcode.STATUS_RESPONSE:
                status_response = interaction_model.StatusResponseMessage.decode(
                    message.application_payload
                )
                print(
                    f"Received Status Response on {message.session_id}/{message.exchange_id} ack {message.acknowledged_message_counter}: {status_response.Status!r}"
                )

                if exchange.pending_payloads:
                    if status_response.Status == interaction_model.StatusCode.SUCCESS:
                        exchange.send(exchange.pending_payloads.pop(0))
                    else:
                        exchange.pending_payloads.clear()
            else:
                print(message)
                print("application payload", message.application_payload.hex(" "))
        else:
            print("Unknown protocol", message.protocol_id, message.protocol_opcode)
        print()

        self.nonvolatile.commit()
        # TODO: Rollback on error?



import enum

from circuitmatter.data_model import (
    Cluster,
    Enum8,
    Map16,
    NumberAttribute,
    OctetStringAttribute,
    BoolAttribute,
    EnumAttribute,
    ListAttribute,
    BitmapAttribute,
)
from circuitmatter import tlv


class ThreadCapabilitiesBitmap(Map16):
    IS_BORDER_ROUTER_CAPABLE = 1 << 0
    IS_ROUTER_CAPABLE = 1 << 1
    IS_SLEEPY_END_DEVICE_CAPABLE = 1 << 2
    IS_FULL_THREAD_DEVICE = 1 << 3
    IS_SYNCHRONIZED_SLEEPY_END_DEVICE_CAPABLE = 1 << 4


class NetworkCommissioningCluster(Cluster):
    CLUSTER_ID = 0x0031

    class FeatureBitmap(enum.IntFlag):
        WIFI_NETWORK_INTERFACE = 0b001
        THREAD_NETWORK_INTERFACE = 0b010
        ETHERNET_NETWORK_INTERFACE = 0b100

    class WifiBandEnum(Enum8):
        BAND_2G4 = 0
        BAND_3G65 = 1
        BAND_5G = 2
        BAND_6G = 3
        BAND_60G = 4
        BAND_1G = 5

    class NetworkCommissioningStatus(Enum8):
        SUCCESS = 0
        """Ok, no error"""

        OUT_OF_RANGE = 1
        """Value Outside Range"""

        BOUNDS_EXCEEDED = 2
        """A collection would exceed its size limit"""

        NETWORK_ID_NOT_FOUND = 3
        """The NetworkID is not among the collection of added networks"""

        DUPLICATE_NETWORK_ID = 4
        """The NetworkID is already among the collection of added networks"""

        NETWORK_NOT_FOUND = 5
        """Cannot find AP: SSID Not found"""

        REGULATORY_ERROR = 6
        """Cannot find AP: Mismatch on band/channels/regulatory domain / 2.4GHz vs 5GHz"""

        AUTH_FAILURE = 7
        """Cannot associate due to authentication failure"""

        UNSUPPORTED_SECURITY = 8
        """Cannot associate due to unsupported security mode"""

        OTHER_CONNECTION_FAILURE = 9
        """Other association failure"""

        IPV6_FAILED = 10
        """Failure to generate an IPv6 address"""

        IP_BIND_FAILED = 11
        """Failure to bind Wi-Fi <-> IP interfaces"""

        UNKNOWN_ERROR = 12
        """Unknown error"""

    class NetworkInfoStruct(tlv.Structure):
        NetworkID = tlv.OctetStringMember(0, min_length=1, max_length=32)
        Connected = tlv.BoolMember(1)

    max_networks = NumberAttribute(0, signed=False, bits=8, default=1, F_fixed=True)
    networks = ListAttribute(1, NetworkInfoStruct)
    scan_max_time_seconds = NumberAttribute(
        2,
        signed=False,
        bits=8,
        feature=FeatureBitmap.WIFI_NETWORK_INTERFACE
        | FeatureBitmap.THREAD_NETWORK_INTERFACE,
        F_fixed=True,
    )
    connect_max_time_seconds = NumberAttribute(
        3,
        signed=False,
        bits=8,
        feature=FeatureBitmap.WIFI_NETWORK_INTERFACE
        | FeatureBitmap.THREAD_NETWORK_INTERFACE,
        F_fixed=True,
    )
    interface_enabled = BoolAttribute(4, default=True, N_nonvolatile=True)
    last_network_status = EnumAttribute(5, NetworkCommissioningStatus, X_nullable=True)
    last_network_id = OctetStringAttribute(
        6, min_length=1, max_length=32, X_nullable=True
    )
    last_connect_error_value = NumberAttribute(7, signed=True, bits=32, X_nullable=True)
    supported_wifi_bands = ListAttribute(
        8, WifiBandEnum, feature=FeatureBitmap.WIFI_NETWORK_INTERFACE, F_fixed=True
    )
    supported_thread_features = BitmapAttribute(
        9,
        ThreadCapabilitiesBitmap,
        feature=FeatureBitmap.THREAD_NETWORK_INTERFACE,
        F_fixed=True,
    )
    thread_version = NumberAttribute(
        10,
        signed=False,
        bits=16,
        feature=FeatureBitmap.THREAD_NETWORK_INTERFACE,
        F_fixed=True,
    )



from circuitmatter import data_model
from circuitmatter import tlv


class UserLabelCluster(data_model.Cluster):
    CLUSTER_ID = 0x0041
    cluster_revision = 1

    class LabelStruct(tlv.Structure):
        Label = tlv.UTF8StringMember(0, 16, default="")
        Value = tlv.UTF8StringMember(1, 16, default="")

    LabelList = data_model.ListAttribute(
        0x0000, LabelStruct, default=[], N_nonvolatile=True
    )



import enum
import struct

from . import tlv
from .protocol import ProtocolId

from typing import Optional


class ExchangeFlags(enum.IntFlag):
    V = 1 << 4
    SX = 1 << 3
    R = 1 << 2
    A = 1 << 1
    I = 1 << 0  # noqa: E741


class SecurityFlags(enum.IntFlag):
    P = 1 << 7
    C = 1 << 6
    MX = 1 << 5
    # This is actually 2 bits but the top bit is reserved and always zero.
    GROUP = 1 << 0


class Message:
    def __init__(self):
        self.clear()
        self.buffer = None

    def clear(self):
        self.flags: int = 0
        self.session_id: int = 0
        self.security_flags: SecurityFlags = SecurityFlags(0)
        self.message_counter: Optional[int] = None
        self.source_node_id = 0
        self.destination_node_id = 0
        self.secure_session: Optional[bool] = None
        self.payload = None
        self.duplicate: Optional[bool] = None

        # Filled in after the message payload is decrypted.
        self.exchange_flags: ExchangeFlags = ExchangeFlags(0)
        self.exchange_id: Optional[int] = None

        self.protocol_vendor_id = 0
        self.protocol_id = ProtocolId(0)
        self.protocol_opcode: Optional[int] = None

        self.acknowledged_message_counter = None
        self.application_payload = None

        self.source_ipaddress = None

        self.header = None

    def parse_protocol_header(self):
        self.exchange_flags, self.protocol_opcode, self.exchange_id = (
            struct.unpack_from("<BBH", self.payload)
        )

        self.exchange_flags = ExchangeFlags(self.exchange_flags)
        decrypted_offset = 4
        self.protocol_vendor_id = 0
        if self.exchange_flags & ExchangeFlags.V:
            self.protocol_vendor_id = struct.unpack_from(
                "<H", self.payload, decrypted_offset
            )[0]
            decrypted_offset += 2
        protocol_id = struct.unpack_from("<H", self.payload, decrypted_offset)[0]
        decrypted_offset += 2
        self.protocol_id = ProtocolId(protocol_id)
        self.protocol_opcode = self.protocol_id.ProtocolOpcode(self.protocol_opcode)

        self.acknowledged_message_counter = None
        if self.exchange_flags & ExchangeFlags.A:
            self.acknowledged_message_counter = struct.unpack_from(
                "<I", self.payload, decrypted_offset
            )[0]
            decrypted_offset += 4

        self.application_payload = self.payload[decrypted_offset:]

    def decode(self, buffer):
        self.clear()
        self.buffer = buffer
        self.flags, self.session_id, self.security_flags, self.message_counter = (
            struct.unpack_from("<BHBI", buffer)
        )
        self.security_flags = SecurityFlags(self.security_flags)
        offset = 8
        if self.flags & (1 << 2):
            self.source_node_id = struct.unpack_from("<Q", buffer, 8)[0]
            offset += 8
        else:
            self.source_node_id = 0

        dsiz = self.flags & 0b11
        if dsiz == 1:
            self.destination_node_id = struct.unpack_from("<Q", buffer, offset)[0]
            offset += 8
        elif dsiz == 2:
            self.destination_node_id = struct.unpack_from("<H", buffer, offset)[0]
            offset += 2
            self.destination_node_id |= 0xFFFF_FFFF_FFFF_0000

        if (self.flags >> 4) != 0:
            raise RuntimeError("Incorrect version")
        self.secure_session = not (
            not (self.security_flags & SecurityFlags.GROUP) and self.session_id == 0
        )
        self.decrypted = not self.secure_session

        self.header = memoryview(buffer)[:offset]
        self.payload = memoryview(buffer)[offset:]
        self.duplicate = None

    def encode_into(self, buffer, cipher=None):
        if self.buffer is not None:
            buffer[: len(self.buffer)] = self.buffer
            return len(self.buffer)
        offset = 0
        struct.pack_into(
            "<BHBI",
            buffer,
            offset,
            self.flags,
            self.session_id,
            self.security_flags,
            self.message_counter,
        )
        nonce_start = 3
        nonce_end = nonce_start + 1 + 4
        offset += 8
        if self.flags & (1 << 2):
            struct.pack_into("<Q", buffer, offset, self.source_node_id)
            offset += 8
            nonce_end += 8
        if (self.flags & 0b11) != 0:
            if self.destination_node_id > 0xFFFF_FFFF_FFFF_0000:
                struct.pack_into(
                    "<H", buffer, offset, self.destination_node_id & 0xFFFF
                )
                offset += 2
            else:
                struct.pack_into("<Q", buffer, offset, self.destination_node_id)
                offset += 8

        if cipher is not None:
            unencrypted_buffer = memoryview(bytearray(1280))
            unencrypted_offset = 0
        else:
            unencrypted_buffer = buffer
            unencrypted_offset = offset

        struct.pack_into(
            "BBHH",
            unencrypted_buffer,
            unencrypted_offset,
            self.exchange_flags,
            self.protocol_opcode,
            self.exchange_id,
            self.protocol_id,
        )
        unencrypted_offset += 6
        if self.acknowledged_message_counter is not None:
            struct.pack_into(
                "I",
                unencrypted_buffer,
                unencrypted_offset,
                self.acknowledged_message_counter,
            )
            unencrypted_offset += 4

        if self.application_payload is not None:
            if isinstance(self.application_payload, tlv.Structure):
                # Wrap the structure in an anonymous tag.
                unencrypted_buffer[unencrypted_offset] = 0x15
                unencrypted_offset += 1
                unencrypted_offset = self.application_payload.encode_into(
                    unencrypted_buffer, unencrypted_offset
                )
            elif hasattr(self.application_payload, "encode_into"):
                unencrypted_offset = self.application_payload.encode_into(
                    unencrypted_buffer, unencrypted_offset
                )
            else:
                # Skip a copy operation if we're using a separate unencrypted buffer
                if unencrypted_offset == 0:
                    unencrypted_buffer = self.application_payload
                else:
                    unencrypted_buffer[
                        unencrypted_offset : unencrypted_offset
                        + len(self.application_payload)
                    ] = self.application_payload
                unencrypted_offset += len(self.application_payload)

        # print("unencrypted", unencrypted_buffer[:unencrypted_offset].hex(" "))

        # Encrypt the payload
        if cipher is not None:
            # The message may not include the source_node_id so we encode the nonce separately.
            nonce = struct.pack(
                "<BIQ", self.security_flags, self.message_counter, self.source_node_id
            )
            additional = buffer[:offset]
            self.payload = cipher.encrypt(
                nonce, bytes(unencrypted_buffer[:unencrypted_offset]), bytes(additional)
            )
            buffer[offset : offset + len(self.payload)] = self.payload
            offset += len(self.payload)
        else:
            offset = unencrypted_offset

        return offset

    def __str__(self):
        pieces = ["Message:"]
        pieces.append(f"Message Flags: {self.flags}")
        pieces.append(f"Session ID: {self.session_id}")
        pieces.append(f"Security Flags: {self.security_flags}")
        pieces.append(f"Message Counter: {self.message_counter}")

        if self.flags & (1 << 2):
            pieces.append(f"Source Node ID: {self.source_node_id:x}")
        if (self.flags & 0b11) != 0:
            pieces.append(f"Destination Node ID: {self.destination_node_id:x}")
        payload_info = ["Payload: "]
        payload_info.append(f"Exchange Flags: {self.exchange_flags!r}")
        payload_info.append(f"Protocol Opcode: {self.protocol_opcode!r}")
        payload_info.append(f"Exchange ID: {self.exchange_id}")
        if self.protocol_vendor_id:
            payload_info.append(f"Protocol Vendor ID: {self.protocol_vendor_id}")
        payload_info.append(f"Protocol ID: {self.protocol_id!r}")
        if self.acknowledged_message_counter is not None:
            payload_info.append(
                f"Acknowledged Message Counter: {self.acknowledged_message_counter}"
            )
        if self.application_payload is not None:
            application_payload = str(self.application_payload).replace("\n", "\n    ")
            payload_info.append(f"Application Payload: {application_payload}")
        pieces.append("\n    ".join(payload_info))
        return "\n  ".join(pieces)



import enum


class SecureProtocolOpcode(enum.IntEnum):
    MSG_COUNTER_SYNC_REQ = 0x00
    """The Message Counter Synchronization Request message queries the current message counter from a peer to bootstrap replay protection."""

    MSG_COUNTER_SYNC_RSP = 0x01
    """The Message Counter Synchronization Response message provides the current message counter from a peer to bootstrap replay protection."""

    MRP_STANDALONE_ACK = 0x10
    """This message is dedicated for the purpose of sending a stand-alone acknowledgement when there is no other data message available to piggyback an acknowledgement on top of."""

    PBKDF_PARAM_REQUEST = 0x20
    """The request for PBKDF parameters necessary to complete the PASE protocol."""

    PBKDF_PARAM_RESPONSE = 0x21
    """The PBKDF parameters sent in response to PBKDF-ParamRequest during the PASE protocol."""

    PASE_PAKE1 = 0x22
    """The first PAKE message of the PASE protocol."""

    PASE_PAKE2 = 0x23
    """The second PAKE message of the PASE protocol."""

    PASE_PAKE3 = 0x24
    """The third PAKE message of the PASE protocol."""

    CASE_SIGMA1 = 0x30
    """The first message of the CASE protocol."""

    CASE_SIGMA2 = 0x31
    """The second message of the CASE protocol."""

    CASE_SIGMA3 = 0x32
    """The third message of the CASE protocol."""

    CASE_SIGMA2_RESUME = 0x33
    """The second resumption message of the CASE protocol."""

    STATUS_REPORT = 0x40
    """The Status Report message encodes the result of an operation in the Secure Channel as well as other protocols."""

    ICD_CHECK_IN = 0x50
    """The Check-in message notifies a client that the ICD is available for communication."""


class InteractionModelOpcode(enum.IntEnum):
    STATUS_RESPONSE = 0x01
    READ_REQUEST = 0x02
    SUBSCRIBE_REQUEST = 0x03
    SUBSCRIBE_RESPONSE = 0x04
    REPORT_DATA = 0x05
    WRITE_REQUEST = 0x06
    WRITE_RESPONSE = 0x07
    INVOKE_REQUEST = 0x08
    INVOKE_RESPONSE = 0x09
    TIMED_REQUEST = 0x0A


class ProtocolId(enum.IntEnum):
    SECURE_CHANNEL = 0
    INTERACTION_MODEL = 1
    BDX = 2
    USER_DIRECTED_COMMISSIONING = 3
    FOR_TESTING = 4

    def ProtocolOpcode(self, opcode_id: int):
        if self == self.SECURE_CHANNEL:
            return SecureProtocolOpcode(opcode_id)
        elif self == self.INTERACTION_MODEL:
            return InteractionModelOpcode(opcode_id)



from circuitmatter.session import MessageReceptionState


def test_basics():
    """These test the common window behavior"""
    state = MessageReceptionState(123)
    assert state.message_counter == 123

    # Older messages are not ok
    assert state.process_counter(122)

    # The current max is not ok
    assert state.process_counter(123)

    # A current max + 1 is ok
    assert not state.process_counter(124)

    # Old current max isn't
    assert state.process_counter(123)

    # A new value is ok
    assert not state.process_counter(126)

    #
    assert state.process_counter(123)

    assert state.process_counter(124)

    assert not state.process_counter(125)

    assert state.process_counter(124)


def test_window_wrapping():
    """Test wrapping the window data across a rollover"""
    state = MessageReceptionState(123, rollover=True)
    assert state.message_counter == 123

    # Move to the end of the range
    assert not state.process_counter(0xFFFFFFFF)

    # Older is ok when in the window.
    assert not state.process_counter(0xFFFFFFF0)

    # A new value is ok. Window is now 0xFFFFFFF0 to 15
    assert not state.process_counter(16)

    assert state.process_counter(0xFFFFFFF0)

    assert state.process_counter(0xFFFFFFFF)

    assert not state.process_counter(1)

    assert not state.process_counter(0xFFFFFFF8)


def test_unencrypted():
    """These test the common window behavior"""
    state = MessageReceptionState(123, rollover=True, encrypted=False)
    assert state.message_counter == 123

    # Older messages are not ok
    assert state.process_counter(123 - 32)

    # Older messages outside the window are ok
    assert not state.process_counter(123 - 32 - 1)

    assert not state.process_counter(124)


def test_encrypted_no_rollover():
    """These test the common window behavior"""
    state = MessageReceptionState(123, rollover=False, encrypted=True)
    assert state.message_counter == 123

    # Older messages are not ok
    assert state.process_counter(123 - 32)

    # Older messages outside the window are not ok
    assert state.process_counter(123 - 32 - 1)

    # Older messages outside the window are not ok
    assert state.process_counter(0)

    # All newer numbers are ok
    assert not state.process_counter(0xFFFFFFFE)

    # Ok because it is in the window
    assert not state.process_counter(0xFFFFFFFD)

    # All older messages outside the window are not ok
    assert state.process_counter(0)


def test_encrypted_with_rollover():
    """These test the common window behavior"""
    state = MessageReceptionState(123, rollover=True, encrypted=True)
    assert state.message_counter == 123

    # Older messages are not ok
    assert state.process_counter(123 - 32)

    # Older messages outside the window are not ok
    assert state.process_counter(123 - 32 - 1)

    # Older messages outside the window are not ok
    assert state.process_counter(0)

    # Numbers wrapped back within the 2**31 window are not ok
    assert state.process_counter(0xFFFFFFFE)

    assert not state.process_counter(0x80000000)

    assert not state.process_counter(0xFFFFFFFE)

    # Ok because it is in the window
    assert not state.process_counter(0xFFFFFFFD)

    # All older messages outside the window are not ok
    assert state.process_counter(0xFFFFFFFE - 32 - 32)

    assert state.process_counter(0xFFFFFFFE - 0x80000000)

    # It is ok to wrap back around outside the 2**31 window.
    assert not state.process_counter(0xFFFFFFFE - 0x80000000 - 1)



from circuitmatter.session import StatusReport, GeneralCode
from circuitmatter.protocol import ProtocolId


def test_example1():
    report = StatusReport()
    report.general_code = GeneralCode.FAILURE
    report.protocol_id = ProtocolId.BDX
    report.protocol_code = 0x52

    buffer = bytearray(10)
    assert report.encode_into(buffer) == 8
    assert buffer[:8] == bytes.fromhex("01 00 02 00 00 00 52 00")


def test_example1_decode():
    report = StatusReport()
    report.decode(bytes.fromhex("01 00 02 00 00 00 52 00"))

    assert report.general_code == GeneralCode.FAILURE
    assert report.protocol_id == ProtocolId.BDX
    assert report.protocol_code == 0x52


def test_example2():
    report = StatusReport()
    report.general_code = GeneralCode.SUCCESS
    report.protocol_id = 0xFFF1AABB
    report.protocol_code = 0

    buffer = bytearray(10)
    assert report.encode_into(buffer) == 8
    assert buffer[:8] == bytes.fromhex("00 00 BB AA F1 FF 00 00")


def test_example2_decode():
    report = StatusReport()
    report.decode(bytes.fromhex("00 00 BB AA F1 FF 00 00"))

    assert report.general_code == GeneralCode.SUCCESS
    assert report.protocol_id == 0xFFF1AABB
    assert report.protocol_code == 0


def test_protocol_data_example():
    report = StatusReport()
    report.general_code = GeneralCode.FAILURE
    report.protocol_id = 0xFFF1AABB
    report.protocol_code = 9921
    report.protocol_data = [0x55, 0x66, 0xEE, 0xFF]

    buffer = bytearray(20)
    assert report.encode_into(buffer) == 12
    assert buffer[:12] == bytes.fromhex("01 00 BB AA F1 FF C1 26 55 66 EE FF")


def test_protocol_data_example_decode():
    report = StatusReport()
    report.decode(bytes.fromhex("01 00 BB AA F1 FF C1 26 55 66 EE FF"))

    assert report.general_code == GeneralCode.FAILURE
    assert report.protocol_id == 0xFFF1AABB
    assert report.protocol_code == 9921
    assert report.protocol_data == b"\x55\x66\xee\xff"



from circuitmatter.data_model import (
    Cluster,
    Command,
    EndpointNumber,
    Enum8,
    GroupId,
    NumberAttribute,
    ListAttribute,
    List,
)
from circuitmatter import tlv


class GroupKeySetSecurityPolicyEnum(Enum8):
    TRUST_FIRST = 0
    CACHE_AND_SYNC = 1


class GroupKeyMulticastPolicyEnum(Enum8):
    PER_GROUP_ID = 0
    ALL_NODES = 1


class GroupKeySetStruct(tlv.Structure):
    GroupKeySetID = tlv.IntMember(0, signed=False, octets=2)
    GroupKeySecurityPolicy = tlv.EnumMember(1, GroupKeySetSecurityPolicyEnum)
    EpochKey0 = tlv.OctetStringMember(2, 16, nullable=True)
    EpochStartTime0 = tlv.IntMember(3, signed=False, octets=8, nullable=True)
    EpochKey1 = tlv.OctetStringMember(4, 16, nullable=True)
    EpochStartTime1 = tlv.IntMember(5, signed=False, octets=8, nullable=True)
    EpochKey2 = tlv.OctetStringMember(6, 16, nullable=True)
    EpochStartTime2 = tlv.IntMember(7, signed=False, octets=8, nullable=True)
    GroupKeyMulticastPolicy = tlv.EnumMember(
        8, GroupKeyMulticastPolicyEnum, nullable=True
    )


class GroupKeyManagementCluster(Cluster):
    CLUSTER_ID = 0x3F

    class GroupKeyMapStruct(tlv.Structure):
        GroupId = GroupId(1)
        GroupKeySetID = tlv.IntMember(2, signed=False, octets=2, minimum=1)

    class GroupInfoMapStruct(tlv.Structure):
        GroupId = GroupId(1)
        Endpoints = List(2, EndpointNumber())
        GroupName = tlv.UTF8StringMember(3, max_length=16)

    class KeySetWrite(tlv.Structure):
        GroupKeySet = tlv.StructMember(0, GroupKeySetStruct)

    group_key_map = ListAttribute(0, GroupKeyMapStruct, default=[], N_nonvolatile=True)
    group_table = ListAttribute(1, GroupInfoMapStruct, default=[])
    max_groups_per_fabric = NumberAttribute(2, signed=False, bits=16, default=0)
    max_group_keys_per_fabric = NumberAttribute(3, signed=False, bits=16, default=1)

    key_set_write = Command(0, KeySetWrite, None, None)
